# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

# Docker image for running Cebes locally
LOCAL_HTTP_SERVER_REPO = 'cebesio/http-server'
LOCAL_HTTP_SERVER_TAG = '0.10.0'

LOCAL_REPOSITORY_REPO = 'cebesio/pipeline-repo'
LOCAL_REPOSITORY_TAG = '0.10.0'



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import enum

from pycebes.core import stages
from pycebes.internal.helpers import require
from pycebes.internal.implicits import get_default_pipeline


def _create_stage(stage_cls, stage_name=None, **kwargs):
    """
    Internal helper for creating a stage and add it into the current default pipeline
    """
    ppl = get_default_pipeline()

    # invent a name for this stage
    if stage_name is None:
        name_template = '{}_{{}}'.format(stage_cls.__name__.lower())
        idx = 0
        stage_name = name_template.format(idx)
        while ppl.__contains__(stage_name):
            idx += 1
            stage_name = name_template.format(idx)

    inputs = {}
    for k, v in kwargs.items():
        if isinstance(v, stages.Stage):
            # convenience: users can specify a slot, we take the default output slot
            v = v.default_output_slot_descriptor
        if isinstance(v, stages.SlotDescriptor):
            # if v is a slot descriptor, make sure its parent is in the pipeline
            ppl.add(v.parent)
        inputs[k] = v

    return ppl.add(stage_cls(name=stage_name, **inputs))


class PlaceholderTypes(enum.Enum):
    VALUE = 0
    DATAFRAME = 1
    COLUMN = 2


def placeholder(placeholder_type=PlaceholderTypes.DATAFRAME, value_type=None, name=None):
    """

    :param placeholder_type:
    :param value_type:
    :param name:
    :return:
    """
    if placeholder_type == PlaceholderTypes.VALUE:
        return _create_stage(stages.ValuePlaceholder, name, value_type=value_type)
    else:
        placeholders = {PlaceholderTypes.DATAFRAME: stages.DataframePlaceholder,
                        PlaceholderTypes.COLUMN: stages.ColumnPlaceholder}
        require(placeholder_type in placeholders, 'Invalid placeholder type: {}'.format(placeholder_type))
        return _create_stage(placeholders[placeholder_type], name)


"""
ETL stages
"""


def drop(df, col_names, name=None):
    """

    :param df:
    :param col_names:
    :param name:
    :return:
    :rtype: stages.Drop
    """
    return _create_stage(stages.Drop, name, input_df=df, col_names=col_names)


"""
Feature extractors
"""


def index_to_string(df, input_col, output_col, name=None):
    """

    :param df:
    :param input_col:
    :param output_col:
    :param name:
    :return:
    :rtype: stages.IndexToString
    """
    return _create_stage(stages.IndexToString, name, input_df=df, input_col=input_col, output_col=output_col)


def string_indexer(df, input_col, output_col, name=None):
    """

    :param df:
    :param input_col:
    :param output_col:
    :param name:
    :return:
    :rtype: stages.StringIndexer
    """
    return _create_stage(stages.StringIndexer, name, input_df=df, input_col=input_col, output_col=output_col)


def vector_assembler(df, input_cols, output_col='output', name=None):
    """

    :param df:
    :param input_cols: list of column names
    :param output_col: name of the output column
    :param name: name of this stage. Will be automatically generated if not provided
    :return:
    :rtype: stages.VectorAssembler
    """
    return _create_stage(stages.VectorAssembler, name, input_df=df, input_cols=input_cols, output_col=output_col)


"""
ML stages
"""


def linear_regression(df, features_col, label_col, prediction_col='prediction',
                      aggregation_depth=2, elastic_net_param=0.0, fit_intercept=True,
                      max_iter=10, reg_param=0.0, standardization=True,
                      tolerance=1e-6, weight_col=None, solver='auto', name=None):
    """

    :param df:
    :param features_col:
    :param label_col:
    :param prediction_col:
    :param aggregation_depth:
    :param elastic_net_param:
    :param fit_intercept:
    :param max_iter:
    :param reg_param:
    :param standardization:
    :param tolerance:
    :param weight_col:
    :param solver:
    :param name:
    :return:
    :rtype: stages.LinearRegression
    """
    return _create_stage(stages.LinearRegression, name, input_df=df,
                         features_col=features_col, label_col=label_col, prediction_col=prediction_col,
                         aggregation_depth=aggregation_depth, elastic_net_param=elastic_net_param,
                         fit_intercept=fit_intercept, max_iter=max_iter, reg_param=reg_param,
                         standardization=standardization, tolerance=tolerance,
                         weight_col=weight_col, solver=solver)


"""
Evaluators
"""


def _regression_evaluator(df, metric_name, label_col, prediction_col='prediction', name=None):
    return _create_stage(stages.RegressionEvaluator, name,
                         input_df=df, metric_name=metric_name,
                         label_col=label_col, prediction_col=prediction_col)


def rmse(df, label_col, prediction_col='prediction', name=None):
    """

    :param df:
    :param label_col:
    :param prediction_col:
    :param name:
    :return:
    """
    return _regression_evaluator(df, 'rmse', label_col, prediction_col, name)


def mse(df, label_col, prediction_col='prediction', name=None):
    """

    :param df:
    :param label_col:
    :param prediction_col:
    :param name:
    :return:
    """
    return _regression_evaluator(df, 'mse', label_col, prediction_col, name)


def r2(df, label_col, prediction_col='prediction', name=None):
    """

    :param df:
    :param label_col:
    :param prediction_col:
    :param name:
    :return:
    """
    return _regression_evaluator(df, 'r2', label_col, prediction_col, name)


def mae(df, label_col, prediction_col='prediction', name=None):
    """

    :param df:
    :param label_col:
    :param prediction_col:
    :param name:
    :return:
    """
    return _regression_evaluator(df, 'mae', label_col, prediction_col, name)



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.
from pycebes.core.dataframe import Dataframe
from pycebes.core.functions import *
from pycebes.core.pipeline import Pipeline
from pycebes.core.pipeline_api import *
from pycebes.core.schema import StorageTypes, VariableTypes
from pycebes.core.session import Session, ReadOptions, CsvReadOptions, JsonReadOptions, ParquetReadOptions
from pycebes.internal.implicits import get_default_pipeline, get_default_session



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import pandas as pd
import six
import warnings
from pycebes.core.schema import Schema
from pycebes.internal.serializer import from_json


@six.python_2_unicode_compatible
class DataSample(object):
    """
    A sample of data, with a proper schema
    """

    def __init__(self, schema, data):
        """
        
        :type schema: Schema
        :param data: list of list. Each list is a column
        """
        if len(schema) != len(data):
            raise ValueError('Inconsistent data and schema: '
                             '{} fields in schema with {} data columns'.format(len(schema), len(data)))
        self.schema = schema
        self.data = data

    @property
    def columns(self):
        """
        Return a list of column names in this ``DataSample``
        """
        return self.schema.columns

    def __repr__(self):
        return '{}(schema={!r})'.format(self.__class__.__name__, self.schema)

    def to_pandas(self, raise_if_error=False):
        """
        Return a pandas DataFrame representation of this sample

        :param raise_if_error: whether to raise exception when there is a type-cast error
        :rtype: pd.DataFrame
        """
        if len(self.data) == 0:
            return pd.DataFrame()

        data = []
        n_rows = len(self.data[0])
        for i in range(n_rows):
            data.append([c[i] for c in self.data])

        col_names = self.schema.columns
        df = pd.DataFrame(columns=col_names, data=data)

        if len(set(col_names)) < len(col_names):
            warnings.warn('DataSample has duplicated column names. Type casting will not be performed')
        else:
            for f in self.schema.fields:
                # dict (i.e. Map) doesn't work with pandas astype()
                if f.storage_type.python_type is not dict:
                    df[f.name] = df[f.name].astype(dtype=f.storage_type.python_type,
                                                   errors='raise' if raise_if_error else 'ignore')
        return df

    @classmethod
    def from_json(cls, js_data):
        """
        Parse the JSON result from the server

        :param js_data: a dict with a key 'data' for the data part, and 'schema' for the data schema
        :rtype: DataSample
        """
        if 'schema' not in js_data or 'data' not in js_data:
            raise ValueError('Invalid JSON: {}'.format(js_data))

        schema = Schema.from_json(js_data['schema'])
        cols = []
        for c in js_data['data']:
            cols.append([from_json(x) for x in c])
        return DataSample(schema=schema, data=cols)



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import base64
import fcntl
import getpass
import json
import os
import tempfile

import pandas as pd
import six

from pycebes.core.client import Client
from pycebes.core.dataframe import Dataframe
from pycebes.core.pipeline import Model, Pipeline
from pycebes.internal import docker_helpers
from pycebes.internal import responses
from pycebes.internal.helpers import require, get_logger
from pycebes.internal.implicits import get_session_stack

_logger = get_logger(__name__)


@six.python_2_unicode_compatible
class Session(object):
    """
    Construct a new `Session` to the server at the given host and port, with the given user name and password.

    # Arguments
    host (str): Hostname of the Cebes server.
        If `None` (default), cebes will try to launch a new
        docker container with a suitable version of Cebes server in it. Note that it requires you have
        a working docker daemon on your machine.
        Otherwise a string containing the host name or IP address of the Cebes server you want to connect to.
    port (int): The port on which Cebes server is listening. Ignored when ``host=None``.
    user_name (str): Username to log in to Cebes server
    password (str): Password of the user to log in to Cebes server
    interactive (bool): whether this is an interactive session,
        in which case some diagnosis logs will be printed to stdout.
    """

    def __init__(self, host=None, port=21000, user_name='', password='', interactive=True):
        """Construct a Session object. See class docstring for parameters."""
        # local Spark
        self.cebes_container = None
        self.repository_container = None
        if host is None:
            self.cebes_container = docker_helpers.get_cebes_http_server_container()
            host = 'localhost'
            port = self.cebes_container.cebes_port
            _logger.info('Connecting to Cebes container {}'.format(self.cebes_container))
            _logger.info('Spark UI can be accessed at http://localhost:{}'.format(self.cebes_container.spark_port))

        self._client = Client(host=host, port=port, user_name=user_name,
                              password=password, interactive=interactive)

        # the first session created
        session_stack = get_session_stack()
        if session_stack.get_default() is None:
            session_stack.stack.append(self)

    def __repr__(self):
        return '{}(host={!r},port={!r},user_name={!r},api_version={!r})'.format(
            self.__class__.__name__, self._client.host, self._client.port,
            self._client.user_name, self._client.api_version)

    @property
    def client(self):
        """
        Return the client which can be used to send requests to server

        # Returns
        Client: the client object
        """
        return self._client

    @property
    def dataframe(self):
        """
        Return a helper for working with tagged and cached #Dataframe

        # Returns
        _TagHelper:
        """
        return _TagHelper(client=self._client, cmd_prefix='df',
                          object_class=Dataframe, response_class=responses.TaggedDataframeResponse)

    @property
    def model(self):
        """
        Return a helper for working with tagged and cached #Model

        # Returns
        _TagHelper:
        """
        return _TagHelper(client=self._client, cmd_prefix='model',
                          object_class=Model, response_class=responses.TaggedModelResponse)

    @property
    def pipeline(self):
        """
        Return a helper for working with tagged and cached #Pipeline

        # Returns
        _PipelineHelper:
        """
        return _PipelineHelper(client=self._client, cebes_http_container=self.cebes_container,
                               local_repo=self.repository_container)

    def as_default(self):
        """
        Returns a context manager that makes this object the default session.
        
        Use with the `with` keyword to specify that all remote calls to server
        should be executed in this session.
        
        ```python
            sess = cb.Session()
            with sess.as_default():
                ....
        ```

        To get the current default session, use `get_default_session`.
        
        > The default session is a property of the current thread. If you
        > create a new thread, and wish to use the default session in that
        > thread, you must explicitly add a `with sess.as_default():` in that
        > thread's function.
        
        # Returns
          A context manager using this session as the default session.
        """
        return get_session_stack().get_controller(self)

    def start_repository_container(self, host_port=None):
        """
        Start a local docker container running Cebes pipeline repository,
        with the repository listening on the host at the given port.

        If `host_port` is None, a new port will be automatically allocated
        by Docker. Therefore, to maintain consistency with your pipeline tags,
        it is recommended to specify a high port, e.g. 35000 or 36000.

        If one repository was started for this Session already, it will be returned.
        """
        if self.repository_container is None:
            self.repository_container = docker_helpers.get_cebes_repository_container(
                host_port=host_port)
        _logger.info('Pipeline repository started on port {}'.format(self.repository_container.cebes_port))
        return self.repository_container

    def stop_repository_container(self):
        """Stop the local pipeline repository if it is running.
        Do nothing if there is no local repository associated with this Session."""
        if self.repository_container is not None:
            self.repository_container.shutdown()
            self.repository_container = None

    def close(self):
        """
        Close this session. Will stop the Cebes container if this session was
        created against a local Cebes container. Otherwise it is a no-op.
        """
        if self.cebes_container is not None:
            self.cebes_container.shutdown()
            self.cebes_container = None
        self.stop_repository_container()

    """
    Storage APIs
    """

    def _read(self, request):
        """
        Read a Dataframe from the given request

        # Returns
        Dataframe:
        """
        return Dataframe.from_json(self._client.post_and_wait('storage/read', data=request))

    @staticmethod
    def _verify_data_format(fmt='csv', options=None):
        """
        Helper to verify the data format and options
        Return the JSON representation of the given options

        # Arguments
        options (ReadOptions):
        """
        valid_fmts = ['csv', 'json', 'parquet', 'orc', 'text']
        fmt = fmt.lower()
        require(fmt in valid_fmts, 'Unrecognized data format: {}. '
                                   'Supported values are: {}'.format(fmt, ', '.join(valid_fmts)))
        if options is not None:
            if fmt == valid_fmts[0]:
                require(isinstance(options, CsvReadOptions),
                        'options must be a {} object. Got {!r}'.format(CsvReadOptions.__name__, options))
            elif fmt == valid_fmts[1]:
                require(isinstance(options, JsonReadOptions),
                        'options must be a {} object. Got {!r}'.format(JsonReadOptions.__name__, options))
            elif fmt == valid_fmts[2]:
                require(isinstance(options, ParquetReadOptions),
                        'options must be a {} object. Got {!r}'.format(ParquetReadOptions.__name__, options))
            else:
                raise ValueError('options must be None when fmt={}'.format(fmt))
            return options.to_json()
        return {}

    def read_jdbc(self, url, table_name, user_name='', password=''):
        """
        Read a Dataframe from a JDBC table

        # Arguments
        url (str): URL to the JDBC server
        table_name (str): name of the table
        user_name (str): JDBC user name
        password (str): JDBC password

        # Returns
        Dataframe: the Cebes Dataframe object created from the data source
        """
        return self._read({'jdbc': {'url': url, 'tableName': table_name,
                                    'userName': user_name,
                                    'passwordBase64': base64.urlsafe_b64encode(password)}})

    def read_hive(self, table_name=''):
        """
        Read a Dataframe from Hive table of the given name

        # Arguments
        table_name (str): name of the Hive table to read data from

        # Returns
        Dataframe: The Cebes Dataframe object created from Hive table
        """
        return self._read({'hive': {'tableName': table_name}})

    def read_s3(self, bucket, key, access_key, secret_key, region=None, fmt='csv', options=None):
        """
        Read a Dataframe from files stored in Amazon S3.

        # Arguments
        bucket (str): name of the S3 bucket
        key (str): path to the file(s) on S3
        access_key (str): Amazon S3 access key
        secret_key (str): Amazon S3 secret key
        region (str): S3 region, if needed.
        fmt (str): format of the files, can be `csv`, `json`, `orc`, `parquet`, `text`
        options: Additional options that dictate how the files are going to be read.
            If specified, this can be:

            - #CsvReadOptions when `fmt='csv'`,
            - #JsonReadOptions when `fmt='json'`, or
            - #ParquetReadOptions when `fmt='parquet'`

         Other formats do not need additional options

        # Returns
        Dataframe: the Cebes Dataframe object created from the data source
        """
        options_dict = Session._verify_data_format(fmt, options)
        s3_options = {'accessKey': access_key, 'secretKey': secret_key,
                      'bucketName': bucket, 'key': key,
                      'format': fmt}
        if region:
            s3_options['regionName'] = region
        return self._read({'s3': s3_options, 'readOptions': options_dict})

    def read_hdfs(self, path, server=None, fmt='csv', options=None):
        """
        Load a dataset from HDFS.

        # Arguments
        path (str): path to the files on HDFS, e.g. `/data/dataset1`
        server (str): Host name and port, e.g. `hdfs://server:9000`
        fmt (str): format of the files, can be `csv`, `json`, `orc`, `parquet`, `text`
        options: Additional options that dictate how the files are going to be read.
            If specified, this can be:

            - #CsvReadOptions when `fmt='csv'`,
            - #JsonReadOptions when `fmt='json'`, or
            - #ParquetReadOptions when `fmt='parquet'`

         Other formats do not need additional options

        # Returns
        Dataframe: the Cebes Dataframe object created from the data source
        """
        options_dict = Session._verify_data_format(fmt, options)
        hdfs_options = {'path': path, 'format': fmt}
        if server:
            hdfs_options['uri'] = server
        return self._read({'hdfs': hdfs_options, 'readOptions': options_dict})

    def read_local(self, path, fmt='csv', options=None):
        """
        Upload a file from the local machine to the server, and create a :class:`Dataframe` out of it.

        # Arguments
        path (str): path to the local file
        fmt (str): format of the file, can be `csv`, `json`, `orc`, `parquet`, `text`
        options: Additional options that dictate how the files are going to be read.
            If specified, this can be:

            - #CsvReadOptions when `fmt='csv'`,
            - #JsonReadOptions when `fmt='json'`, or
            - #ParquetReadOptions when `fmt='parquet'`

         Other formats do not need additional options

        # Returns
        Dataframe: the Cebes Dataframe object created from the data source
        """
        options_dict = Session._verify_data_format(fmt=fmt, options=options)
        server_path = self._client.upload(path)['path']
        return self._read({'localFs': {'path': server_path, 'format': fmt}, 'readOptions': options_dict})

    def read_csv(self, path, options=None):
        """
        Upload a local CSV file to the server, and create a Dataframe out of it.

        # Arguments
        path (str): path to the local CSV file
        options (CsvReadOptions): Additional options that dictate how the files are going to be read.
            Must be either None or a :class:`CsvReadOptions` object

        # Returns
        Dataframe: the Cebes Dataframe object created from the data source
        """
        return self.read_local(path=path, fmt='csv', options=options)

    def read_json(self, path, options=None):
        """
        Upload a local JSON file to the server, and create a Dataframe out of it.

        # Arguments
        path (str): path to the local JSON file
        options (JsonReadOptions): Additional options that dictate how the files are going to be read.
            Must be either None or a :class:`JsonReadOptions` object

        # Returns
        Dataframe: the Cebes Dataframe object created from the data source
        """
        return self.read_local(path=path, fmt='json', options=options)

    def from_pandas(self, df):
        """
        Upload the given `pandas` DataFrame to the server and create a Cebes Dataframe out of it.
        Types are preserved on a best-efforts basis.

        # Arguments
        df (pd.DataFrame): a pandas DataFrame object

        # Returns
        Dataframe: the Cebes Dataframe created from the data source
        """
        require(isinstance(df, pd.DataFrame), 'Must be a pandas DataFrame object. Got {}'.format(type(df)))
        with tempfile.NamedTemporaryFile('w', prefix='cebes', delete=False) as f:
            df.to_csv(path_or_buf=f, index=False, sep=',', quotechar='"', escapechar='\\', header=True,
                      na_rep='', date_format='yyyy-MM-dd\'T\'HH:mm:ss.SSSZZ')
            file_name = f.name

        csv_options = CsvReadOptions(infer_schema=True,
                                     sep=',', quote='"', escape='\\', header=True,
                                     null_value='', date_format='yyyy-MM-dd\'T\'HH:mm:ss.SSSZZ',
                                     timestamp_format='yyyy-MM-dd\'T\'HH:mm:ss.SSSZZ')
        cebes_df = self.read_csv(file_name, csv_options)
        try:
            os.remove(file_name)
        except IOError:
            pass
        return cebes_df

    def load_test_datasets(self):
        """
        Load test datasets that come by default in Cebes server.

        # Returns
        dict: a dict of datasets, currently has only one element:
            `{'cylinder_bands': Dataframe}`
        """
        response = self._client.post_and_wait('test/loaddata', data={'datasets': ['cylinder_bands']})
        return {'cylinder_bands': Dataframe.from_json(response['dataframes'][0])}


########################################################################

########################################################################


class _TagHelper(object):
    """
    Helper providing tag-related commands
    """

    def __init__(self, client, cmd_prefix='df', object_class=Dataframe,
                 response_class=responses.TaggedDataframeResponse):
        """

        :param client:
        :param cmd_prefix:
        :param object_class:
        :type object_class: Type
        :param response_class:
        :type response_class: Type
        """
        self._object_cls = object_class
        self._client = client
        self._cmd_prefix = cmd_prefix
        self._response_class = response_class

    def get(self, identifier):
        """
        Get the object from the given identifier, which can be a tag or a UUID

        :param identifier: either a tag or an ID of the object to be retrieved.
        :type identifier: six.text_type
        """
        return self._object_cls.from_json(self._client.post_and_wait('{}/get'.format(self._cmd_prefix), identifier))

    def tag(self, obj, tag):
        """
        Add the given tag to the given object, return the object itself

        :param obj: the object to be tagged
        :param tag: new tag for the object
        :type tag: six.text_type
        :return: the object itself if success
        """
        require(isinstance(obj, self._object_cls), 'Unsupported object of type {}'.format(type(obj)))
        self._client.post_and_wait('{}/tagadd'.format(self._cmd_prefix), {'tag': tag, 'objectId': obj.id})
        return obj

    def untag(self, tag):
        """
        Untag the object of the given tag. Note that if the object
        has more than 1 tag, it can still be accessed using other tags.

        :param tag: the tag of the object to be removed
        :type tag: six.text_type
        :return: the object itself if success
        """
        return self._object_cls.from_json(self._client.post_and_wait(
            '{}/tagdelete'.format(self._cmd_prefix), {'tag': tag}))

    def list(self, pattern=None, max_count=100):
        """
        Get the list of tagged objects.

        :param pattern: a pattern string to match the tags.
            Simple wildcards are supported: use ``?`` to match zero or one character,
            ``*`` to match zero or more characters.
        :type pattern: six.text_type
        :param max_count: maximum number of entries to be returned
        :rtype: responses._TaggedResponse
        """
        data = {'maxCount': max_count}
        if pattern is not None:
            data['pattern'] = pattern
        return self._response_class(self._client.post_and_wait('{}/tags'.format(self._cmd_prefix), data))


class _PipelineHelper(_TagHelper):
    """
    Helper for pipeline commands
    """

    def __init__(self, client, cebes_http_container, local_repo):
        super(_PipelineHelper, self).__init__(
            client=client, cmd_prefix='pipeline',
            object_class=Pipeline, response_class=responses.TaggedPipelineResponse)
        self._cebes_http_container = cebes_http_container
        self._local_repo = local_repo

    def login(self, host=None, port=80, username='', password=None):
        """
        Login to the repository at the given host and port, with the given name and password

        If username is non-empty and password is None, pycebes will ask for a password.
        This way the password will not be stored in plain-text in the (I)Python interpreter history.
        """
        if username and password is None:
            password = getpass.getpass('{}\' password: '.format(username))
        data = {'userName': username, 'passwordHash': password or ''}

        if host:
            data.update({'host': host, 'port': port})
        else:
            # if host is not provided, try to get reasonable defaults
            default_host, default_port = self._get_default_repo_host_and_port()
            if default_host:
                data.update({'host': default_host, 'port': default_port})

        result = self._client.post_and_wait('pipeline/login', data)

        self._update_repo_creds(host=result['host'], port=result['port'],
                                token=result['token'])
        _logger.info('Logged into repository {}:{}'.format(result['host'], result['port']))

    def push(self, ppl):
        """
        Push the pipeline to the repository
        """
        data = self._get_tag_info(ppl)
        result = self._client.post_and_wait('pipeline/push', data)
        _logger.info('Pushed successfully: {}'.format(result))

    def pull(self, tag):
        """
        Pull the pipeline with the given tag, return it as an object

        # Returns
        Pipeline: the pipeline object
        """
        data = self._get_tag_info(tag)
        result = self._client.post_and_wait('pipeline/pull', data)
        _logger.info('Pulled successfully: {}'.format(data['tag']))
        return Pipeline.from_json(result)

    def _get_default_repo_host_and_port(self):
        """Get the default repository host and port"""
        local_cebes_http_server = self._client.host in ('127.0.0.1', 'localhost')
        if local_cebes_http_server and self._local_repo:
            if self._cebes_http_container:
                # both cebes-http-server and cebes-repository are running in Docker containers
                # use docker ip address and default repository port
                return self._local_repo.container_ip, 22000
            else:
                # Cebes-http-server is running locally but not in Docker (e.g. with ./bin/start-cebes.sh)
                # use localhost and mapped port
                return 'localhost', self._local_repo.cebes_port
        return None, None

    def _get_tag_info(self, tag_or_ppl):
        """Get information of the given tag."""
        data = {}
        if isinstance(tag_or_ppl, Pipeline):
            data['identifier'] = tag_or_ppl.id
        else:
            data['identifier'] = six.text_type(tag_or_ppl)

        default_host, default_port = self._get_default_repo_host_and_port()
        if default_host:
            data.update({'defaultRepoHost': default_host, 'defaultRepoPort': default_port})

        tag_info = self._client.post_and_wait('pipeline/taginfo', data)
        repo_port = int(tag_info['port'])
        _logger.info('Using repository {}:{}'.format(tag_info['host'], repo_port))
        token = self._read_repo_creds(tag_info['host'], repo_port)

        data = {'tag': tag_info['tag'], 'host': tag_info['host'], 'port': repo_port}
        if token:
            data['token'] = token
        return data

    def _open_creds_file(self):
        cebes_data_path = os.path.expanduser('~/.cebes')
        os.makedirs(cebes_data_path, mode=0o700, exist_ok=True)
        creds_file_path = os.path.join(cebes_data_path, 'repositories.json')
        try:
            return open(creds_file_path, 'r+')
        except FileNotFoundError:
            return open(creds_file_path, 'a+')

    def _read_repo_creds(self, host, port):
        with self._open_creds_file() as f:
            fcntl.flock(f, fcntl.LOCK_EX)
            try:
                try:
                    repositories = json.load(f)
                except json.JSONDecodeError:
                    repositories = []

                return next((entry.get('token', '') for entry in repositories
                            if entry['host'] == host and entry['port'] == port), '')
            finally:
                fcntl.flock(f, fcntl.LOCK_UN)

    def _update_repo_creds(self, host, port, token):
        with self._open_creds_file() as f:
            fcntl.flock(f, fcntl.LOCK_EX)
            try:
                try:
                    repositories = json.load(f)
                except json.JSONDecodeError:
                    repositories = []

                updated_repos = []
                for entry in repositories:
                    if entry['host'] != host or entry['port'] != port:
                        updated_repos.append(entry)
                updated_repos.append({'host': host, 'port': port, 'token': token})

                f.seek(0)
                f.write(json.dumps(updated_repos))
                f.truncate()
            finally:
                fcntl.flock(f, fcntl.LOCK_UN)


########################################################################

########################################################################


class ReadOptions(object):
    PERMISSIVE = 'PERMISSIVE'
    DROPMALFORMED = 'DROPMALFORMED'
    FAILFAST = 'FAILFAST'

    """
    Contain options for read commands
    """

    def __init__(self, **kwargs):
        self.options = kwargs

    def to_json(self):
        """
        Convert into a dict of options, with keys suitable for the server
        Concretely, this function will convert the keys from snake_convention to camelConvention
        """
        d = {}
        for k, v in self.options.items():
            # skip values that are None
            if v is None:
                continue

            # convert the key from snake_convention to camelConvention (for server)
            k_str = ''
            i = 0
            while i < len(k):
                if k[i] == '_' and i < len(k) - 1:
                    k_str += k[i + 1].upper()
                    i += 2
                else:
                    k_str += k[i]
                    i += 1

            # convert the value into its string representation
            if isinstance(v, bool):
                v_str = 'true' if v else 'false'
            else:
                v_str = '{}'.format(v)

            d[k_str] = v_str
        return d


class CsvReadOptions(ReadOptions):
    """
    Options for reading CSV files.

    # Arguments
    sep: sets the single character as a separator for each field and value.
    encoding: decodes the CSV files by the given encoding type
    quote: sets the single character used for escaping quoted values where
        the separator can be part of the value. If you would like to turn off quotations, you need to
        set not `null` but an empty string.
    escape: sets the single character used for escaping quotes inside an already quoted value.
    comment: sets the single character used for skipping lines beginning with this character.
        By default, it is disabled
    header: uses the first line as names of columns.
    infer_schema: infers the input schema automatically from data. It requires one extra pass over the data.
    ignore_leading_white_space: defines whether or not leading whitespaces
        from values being read should be skipped.
    null_value: sets the string representation of a null value.
        This applies to all supported types including the string type.
    nan_value: sets the string representation of a "non-number" value
    positive_inf: sets the string representation of a positive infinity value
    negative_inf: sets the string representation of a negative infinity value
    date_format: sets the string that indicates a date format.
        Custom date formats follow the formats at `java.text.SimpleDateFormat`.
        This applies to date type.
    timestamp_format: sets the string that indicates a timestamp format.
        Custom date formats follow the formats at `java.text.SimpleDateFormat`. This applies to timestamp type.
    max_columns: defines a hard limit of how many columns a record can have
    max_chars_per_column: defines the maximum number of characters allowed
        for any given value being read. By default, it is -1 meaning unlimited length
    max_malformed_log_per_partition: sets the maximum number of malformed rows
        will be logged for each partition. Malformed records beyond this number will be ignored.
    mode: allows a mode for dealing with corrupt records during parsing.
        - #ReadOptions.PERMISSIVE - sets other fields to `null` when it meets a corrupted record.
                When a schema is set by user, it sets `null` for extra fields
        - #ReadOptions.DROPMALFORMED - ignores the whole corrupted records
        - #ReadOptions.FAILFAST - throws an exception when it meets corrupted records

    """

    def __init__(self, sep=',', encoding='UTF-8', quote='"', escape='\\', comment=None, header=False,
                 infer_schema=False, ignore_leading_white_space=False, null_value=None, nan_value='NaN',
                 positive_inf='Inf', negative_inf='-Inf', date_format='yyyy-MM-dd',
                 timestamp_format='yyyy-MM-dd\'T\'HH:mm:ss.SSSZZ', max_columns=20480,
                 max_chars_per_column=-1, max_malformed_log_per_partition=10, mode=ReadOptions.PERMISSIVE):
        """See class docstring for documentation."""

        super(CsvReadOptions, self).__init__(sep=sep, encoding=encoding, quote=quote,
                                             escape=escape, comment=comment, header=header,
                                             infer_schema=infer_schema,
                                             ignore_leading_white_space=ignore_leading_white_space,
                                             null_value=null_value, nan_value=nan_value,
                                             positive_inf=positive_inf, negative_inf=negative_inf,
                                             date_format=date_format, timestamp_format=timestamp_format,
                                             max_columns=max_columns, max_chars_per_column=max_chars_per_column,
                                             max_malformed_log_per_partition=max_malformed_log_per_partition,
                                             mode=mode)


class JsonReadOptions(ReadOptions):
    """
    Options for reading Json files

    # Arguments
    primitives_as_string: infers all primitive values as a string type
    prefers_decimal: infers all floating-point values as a decimal type.
        If the values do not fit in decimal, then it infers them as doubles
    allow_comments: ignores Java/C++ style comment in JSON records
    allow_unquoted_field_names: allows unquoted JSON field names
    allow_single_quotes: allows single quotes in addition to double quotes
    allow_numeric_leading_zeros: allows leading zeros in numbers (e.g. 00012)
    allow_backslash_escaping_any_character: allows accepting quoting of all
        character using backslash quoting mechanism
    mode: allows a mode for dealing with corrupt records during parsing.
        - `ReadOptions.PERMISSIVE` - sets other fields to `null` when it meets a corrupted record.
                When a schema is set by user, it sets `null` for extra fields
        - `ReadOptions.DROPMALFORMED` - ignores the whole corrupted records
        - `ReadOptions.FAILFAST` - throws an exception when it meets corrupted records

    column_name_of_corrupt_record: allows renaming the new field having malformed string
        created by `ReadOptions.PERMISSIVE` mode. This overrides `spark.sql.columnNameOfCorruptRecord`.
    date_format: sets the string that indicates a date format.
        Custom date formats follow the formats at `java.text.SimpleDateFormat`. This applies to date type
    timestamp_format: sets the string that indicates a timestamp format.
        Custom date formats follow the formats at `java.text.SimpleDateFormat`. This applies to timestamp type
    """

    def __init__(self, primitives_as_string=False, prefers_decimal=False, allow_comments=False,
                 allow_unquoted_field_names=False, allow_single_quotes=True, allow_numeric_leading_zeros=False,
                 allow_backslash_escaping_any_character=False, mode=ReadOptions.PERMISSIVE,
                 column_name_of_corrupt_record=None, date_format='yyyy-MM-dd',
                 timestamp_format="yyyy-MM-dd'T'HH:mm:ss.SSSZZ"):
        """See class docstring for documentation."""

        super(JsonReadOptions,
              self).__init__(primitives_as_string=primitives_as_string,
                             prefers_decimal=prefers_decimal, allow_comments=allow_comments,
                             allow_unquoted_field_names=allow_unquoted_field_names,
                             allow_single_quotes=allow_single_quotes,
                             allow_numeric_leading_zeros=allow_numeric_leading_zeros,
                             allow_backslash_escaping_any_character=allow_backslash_escaping_any_character,
                             mode=mode, column_name_of_corrupt_record=column_name_of_corrupt_record,
                             date_format=date_format, timestamp_format=timestamp_format)


class ParquetReadOptions(ReadOptions):
    """
    Options for reading Parquet files

    # Arguments
    merge_schema (bool): sets whether we should merge schemas collected from all Parquet part-files.
    """

    def __init__(self, merge_schema=True):
        """See class docstring for documentation."""
        super(ParquetReadOptions, self).__init__(merge_schema=merge_schema)



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import functools
from collections import namedtuple

import six

from pycebes.core.column import Column
from pycebes.core.dataframe import Dataframe
from pycebes.core.sample import DataSample
from pycebes.internal import serializer
from pycebes.internal.helpers import require
from pycebes.internal.implicits import get_default_session

# _Slot is the internal representation of a slot, belong to the class, not the object
# It is used to keep information about how to communicate with the server
_Slot = namedtuple('_Slot', ('name', 'message_type', 'is_input', 'server_name', 'is_default'))


class MessageType(object):
    VALUE_DEF = 'ValueDef'
    STAGE_OUTPUT_DEF = 'StageOutputDef'
    DATAFRAME_DEF = 'DataframeMessageDef'
    SAMPLE_DEF = 'SampleMessageDef'
    MODEL_DEF = 'ModelMessageDef'
    COLUMN_DEF = 'ColumnDef'

    __all_defs = [VALUE_DEF, STAGE_OUTPUT_DEF, DATAFRAME_DEF, SAMPLE_DEF, MODEL_DEF, COLUMN_DEF]

    def __init__(self, msg_type=VALUE_DEF, value_type=None):
        require(msg_type in self.__all_defs, 'Invalid msg_type: {!r}'.format(msg_type))
        self.msg_type = msg_type
        self.value_type = value_type

    def __repr__(self):
        return '{}(msg_type={!r},value_type={!r})'.format(self.__class__.__name__, self.msg_type, self.value_type)

    def is_valid(self, value):
        """
        Check if the given value is valid with the Message type
        :param value: value to be checked
        """
        if value is None:
            return True

        _types = {self.VALUE_DEF: (int, float, six.text_type, list, tuple, dict),
                  self.DATAFRAME_DEF: Dataframe,
                  self.SAMPLE_DEF: DataSample,
                  self.COLUMN_DEF: Column}

        for k, value_type in _types.items():
            if self.msg_type == k:
                if isinstance(value, SlotDescriptor):
                    # in case  the value is a output slot, its type
                    # must be the same with the type of this message type
                    return value.message_type.msg_type == k
                return isinstance(value, value_type)

        if self.msg_type == self.MODEL_DEF:
            # TODO: implement models
            raise NotImplementedError()

        require(self.msg_type == self.STAGE_OUTPUT_DEF, 'Unrecognized message type: {}'.format(self.msg_type))
        return isinstance(value, SlotDescriptor)

    def to_json(self, value):
        """
        Serialize the given value to JSON format, given the message type
        """
        require(self.is_valid(value), 'Invalid value for message type {!r}: {!r}'.format(self, value))

        returned_msg_type = self.msg_type
        if value is None:
            js = None

        elif isinstance(value, SlotDescriptor):
            # stage output message
            require(value.message_type.msg_type == self.msg_type,
                    'Incompatible message type between input slot of type {!r} and '
                    'output slot {!r}'.format(self, value))
            returned_msg_type = self.STAGE_OUTPUT_DEF
            js = value.to_json()

        elif self.msg_type == self.VALUE_DEF:
            js = serializer.to_json(value, param_type=self.value_type)
        elif self.msg_type == self.STAGE_OUTPUT_DEF:
            js = value.to_json()
        elif self.msg_type == self.DATAFRAME_DEF:
            js = {'dfId': value.id}
        elif self.msg_type == self.SAMPLE_DEF:
            raise NotImplementedError()
        elif self.msg_type == self.MODEL_DEF:
            raise NotImplementedError()
        else:
            require(self.msg_type == self.COLUMN_DEF, 'Unrecognized message type: {}'.format(self.msg_type))
            js = value.to_json()
        return [returned_msg_type, js]

    @classmethod
    def from_json(cls, js_data):
        """
        Read a PipelineMessageDef (from server) to the value

        :param js_data: a list of 2 elelements, see ``to_json()`` for more information
        :return: value read from the JSON result
        """
        require(len(js_data) == 2, 'Invalid JSON data: {!r}'.format(js_data))
        msg_type = js_data[0]
        msg_content = js_data[1]

        if msg_type == MessageType.VALUE_DEF:
            return serializer.from_json(msg_content)
        if msg_type == MessageType.DATAFRAME_DEF:
            return get_default_session().dataframe.get(msg_content['dfId'])
        if msg_type == MessageType.MODEL_DEF:
            return get_default_session().model.get(msg_content['modelId'])
        if msg_type == MessageType.COLUMN_DEF:
            raise NotImplementedError('{}'.format(js_data))
        if msg_type == MessageType.STAGE_OUTPUT_DEF:
            # we don't have enough information to return a SlotDescriptor here
            return None

        raise NotImplementedError('{}'.format(js_data))


class MessageTypes(object):
    VALUE = MessageType(MessageType.VALUE_DEF)
    STAGE_OUTPUT = MessageType(MessageType.STAGE_OUTPUT_DEF)
    DATAFRAME = MessageType(MessageType.DATAFRAME_DEF)
    SAMPLE = MessageType(MessageType.SAMPLE_DEF)
    MODEL = MessageType(MessageType.MODEL_DEF)
    COLUMN = MessageType(MessageType.COLUMN_DEF)

    @classmethod
    def value(cls, value_type=None):
        """Same with VALUE, but user has the chance to specify a custom value type for the value"""
        return MessageType(MessageType.VALUE_DEF, value_type=value_type)


##############################################################################

##############################################################################

class SlotDescriptor(object):
    """
    SlotDescriptor belong to the object, not the class
    It contains the name of the stage it belongs to
    """

    def __init__(self, parent, name='', is_input=True):
        """
        Construct a new SlotDescriptor

        :param parent: the parent Stage object
        :type parent: Stage
        :param name: name of this slot
        :type name: six.text_type
        :param is_input: Whether this is an input slot
        """
        self.parent = parent
        self.name = name
        self.is_input = is_input

    def __repr__(self):
        return '{}(name={!r},is_input={!r})'.format(self.__class__.__name__, self.name, self.is_input)

    @property
    def parent_name(self):
        """Returns the name of the parent stage"""
        return self.parent.get_name()

    @property
    def _parent_slot(self):
        """The ``_Slot`` object correspond to this SlotDescriptor.
        Computed by looking up the parent"""
        return self.parent.slot(self.name, self.is_input)

    @property
    def server_name(self):
        """Return the server name of this slot. Computed by looking up the slot in the parent"""
        return self._parent_slot.server_name

    @property
    def full_name(self):
        """Return the fully qualified name, normally in the format ``<parent name>:<name>``"""
        return '{}:{}'.format(self.parent_name, self.name)

    @property
    def full_server_name(self):
        """Return the fully qualified name, normally in the format ``<parent name>:<server name>``"""
        return '{}:{}'.format(self.parent_name, self.server_name)

    @property
    def message_type(self):
        """Return the message type of this slot.
        Computed by looking up the parent

        :rtype: MessageType
        """
        return self._parent_slot.message_type

    def to_json(self):
        """
        Return the JSON representation of this slot descriptor
        """
        return {'stageName': self.parent.get_input(self.parent.name),
                'outputName': self.server_name}


####################################################################################

####################################################################################


def _get_slots(cls, is_input=True):
    """
    Get the list of slots of the given class

    :type cls: Type
    """
    slots = []
    for parent_class in cls.mro():
        ss = getattr(cls, 'SLOTS', {})
        cls_slots = [p for p in ss.get(parent_class.__name__, []) if p.is_input == is_input]
        slots.extend(cls_slots)
    return slots


def _add_slot(name='', message_type=MessageTypes.VALUE, server_name=None, doc='', is_input=True, is_default=False):
    def decorate(cls):
        class_name = cls.__name__
        if class_name not in cls.SLOTS:
            cls.SLOTS[class_name] = []

        s = _Slot(name, message_type, is_input, name if server_name is None else server_name, is_default=is_default)
        require(next((p for p in _get_slots(cls, is_input) if p.name == s.name), None) is None,
                'Duplicated slot named {} in class {}'.format(s.name, class_name))
        cls.SLOTS[class_name].append(s)

        def prop_get(slot_name, is_input_slot, self):
            slot_desc_attr = '_{}'.format(slot_name)
            slot_desc = getattr(self, slot_desc_attr, None)
            if slot_desc is None:
                slot_desc = SlotDescriptor(self, slot_name, is_input=is_input_slot)
                setattr(self, slot_desc_attr, slot_desc)
            return slot_desc

        prop_doc = '{} slot of type {}: {}'.format('Input' if is_input else 'Output',
                                                   message_type.msg_type, doc)
        setattr(cls, s.name, property(fget=functools.partial(prop_get, s.name, is_input), doc=prop_doc))
        return cls

    return decorate


def input_slot(name='', message_type=MessageTypes.VALUE, server_name=None, doc=''):
    """
    Create an input slot for the Stage

    :param name: name of the slot
    :param message_type: type of the message used to transport the slot value
    :param server_name: optionally, the name of this input on the server. If not specified,
        it is the same with ``name``.
    :param doc: Brief documentation explaining the slot
    """
    return _add_slot(name=name, message_type=message_type,
                     server_name=server_name, doc=doc, is_input=True)


def output_slot(name='', message_type=MessageTypes.VALUE, server_name=None, doc='', is_default=False):
    """
    Create an output slot for the Stage

    :param name: name of the slot
    :param message_type: type of the message used to transport the slot value
    :param server_name: optionally, the name of this input on the server. If not specified,
        it is the same with ``name``.
    :param doc: Brief documentation explaining the slot
    :param is_default: is this default output slot
    """
    return _add_slot(name=name, message_type=message_type, server_name=server_name, doc=doc,
                     is_input=False, is_default=is_default)


####################################################################################

####################################################################################

@input_slot('name', MessageTypes.VALUE)
@six.python_2_unicode_compatible
class Stage(object):
    SLOTS = {}

    def __init__(self, name='stage', **kwargs):
        """
        Initialize the stage

        :param name: name given to this stage
        :param kwargs: values assigned to the input slots
        """
        self._slot_values = {}

        # Name is special. We need this to avoid infinite recursion when looking up other slots
        require(name is not None, 'Stage name should not be None at creation time')
        self._name = SlotDescriptor(self, 'name', True)
        self._set_input(self.name, name)

        for k, v in kwargs.items():
            self._set_input(self.slot_descriptor(k), v)

    def __repr__(self):
        slot_desc = ','.join('{}={!r}'.format(s.name, self.get_input(self.slot_descriptor(s.name)))
                             for s in _get_slots(self.__class__, True))
        return '{}({})'.format(self.__class__.__name__, slot_desc)

    def __dir__(self):
        return dir(type(self)) + [p.name for p in _get_slots(self.__class__, True) + _get_slots(self.__class__, False)]

    def _set_input(self, slot_desc, value):
        """
        Set the value of the given slot descriptor

        This is a private function for a reason: users are not supposed to change the input slots
        of a stage, once it is already created. The only way to update the slots is to use
        the `feeds` argument in the call to Pipeline.run().

        :param slot_desc: The slot descriptor to set the value to
        :type slot_desc: SlotDescriptor
        :param value:
        :return: this instance
        """
        require(isinstance(slot_desc, SlotDescriptor) and slot_desc.parent is self and slot_desc.is_input,
                'Not a slot descriptor or it does not belong to this instance: {}'.format(slot_desc.name))
        if isinstance(value, SlotDescriptor):
            # output slot of another stage
            require((not value.is_input) and value.message_type.msg_type == slot_desc.message_type.msg_type,
                    'Slot {!r} is not an output or of incompatible type for input slot {}'.format(
                        value, slot_desc.name))
        else:
            require(slot_desc.message_type.is_valid(value),
                    'Invalid type of value {!r} for slot {}'.format(value, slot_desc.name))
        self._slot_values[slot_desc.name] = value
        return self

    """
    Public APIs
    """

    @property
    def default_output_slot_descriptor(self):
        """
        Return the default output slot descriptor of this stage.
        This is used to provide better UX for end-user.

        :rtype: SlotDescriptor
        """
        all_output_slots = _get_slots(self.__class__, is_input=False)
        require(len(all_output_slots) > 0,
                'Stage {}(name={}) has no output slot'.format(self.__class__.__name__, self.get_name()))

        if len(all_output_slots) == 1:
            default_slot = all_output_slots[0]
        else:
            default_slot = next((s for s in all_output_slots if s.is_default), None)
            require(default_slot is not None, 'No default output slot defined for stage {}(name={})'.format(
                self.__class__.__name__, self.get_name()))

        return self.slot_descriptor(default_slot.name)

    def slot(self, slot_name='', is_input=True):
        """
        Return the slot with the given name

        :rtype: _Slot
        """
        s = next((s for s in _get_slots(self.__class__, is_input) if s.name == slot_name), None)
        require(s is not None, 'Slot name {} not found in class {}'.format(slot_name, self.__class__.__name__))
        return s

    def slot_descriptor(self, slot_name=''):
        """
        Return the SlotDescriptor with the given name

        Different from :func:`slot`, this function return a SlotDescriptor, which is
        a "pointer" representation of the slot, and attaches to a particular Stage object.
        This also works if user provides slot_name as the server_name of the slot

        The _Slot object returned by :func:`slot` is the representation of the slot, attaches to each Stage class

        :rtype: SlotDescriptor
        """
        slot_desc = getattr(self, slot_name, None)
        if slot_desc is not None and isinstance(slot_desc, SlotDescriptor):
            return slot_desc

        # users might be providing the server name. Look it up!
        for c in dir(self):

            if c == Stage.default_output_slot_descriptor.fget.__name__:
                # skip the default output slot descriptor
                continue

            p = getattr(self, c)
            if isinstance(p, SlotDescriptor) and self.slot(c, is_input=p.is_input).server_name == slot_name:
                return p
        raise ValueError('Slot descriptor {} not found in class {}'.format(slot_name, self.__class__.__name__))

    def get_name(self):
        """Returns the name of this stage, shortcut for ``self.get_input(self.name)``
        """
        return self.get_input(self.name)

    def get_input(self, slot_desc, default=None):
        """
        Get the value of the given slot descriptor

        :param slot_desc:
        :type slot_desc: SlotDescriptor
        :param default: default value to  return if the given slot is not specified
        :return: the value
        """
        require(isinstance(slot_desc, SlotDescriptor) and (slot_desc.parent is self) and slot_desc.is_input,
                'Not a slot descriptor or it does not belong to this instance: {}'.format(slot_desc))
        return self._slot_values.get(slot_desc.name, default)

    def to_json(self):
        """
        Serialize the stage into JSON
        """
        inputs = {}
        for s in _get_slots(self.__class__, True):
            if s.name != 'name':
                # skip the name slot
                slot_desc = self.slot_descriptor(s.name)
                v = self.get_input(slot_desc)
                if v is not None:
                    inputs[s.server_name] = s.message_type.to_json(v)

        return {'name': self.get_input(self.name),
                'stageClass': self.__class__.__name__,
                'inputs': inputs,
                'outputs': {},
                'models': {},
                'schemas': {},
                'newInputs': []}

    @staticmethod
    def _find_subclass(cls, name):
        """Find subclass of cls that has the given name, recursively.
        Return None if cannot find any class for this name

        :type cls: Type
        """
        if cls.__name__ == name:
            return cls

        for c in cls.__subclasses__():
            cc = Stage._find_subclass(c, name)
            if cc is not None:
                return cc
        return None

    @classmethod
    def from_json(cls, js_data):
        """
        Return a Stage from its JSON representation

        :param js_data: JSON data (received from server)
        :rtype: Stage
        """
        stage_class = js_data['stageClass']
        clazz = Stage._find_subclass(cls, stage_class)
        require(clazz is not None, 'Could not find any Stage class named {}'.format(stage_class))

        name = js_data['name']
        inputs = {}
        for k, v in js_data['inputs'].items():
            if k != 'name':
                inputs[k] = MessageType.from_json(v)

        return clazz(name=name, **inputs)


@input_slot('input_df', MessageTypes.DATAFRAME, 'inputDf')
@output_slot('output_df', MessageTypes.DATAFRAME, 'outputDf')
class _UnaryTransformer(Stage):
    pass


@input_slot('left_df', MessageTypes.DATAFRAME, 'leftDf')
@input_slot('right_df', MessageTypes.DATAFRAME, 'rightDf')
@output_slot('output_df', MessageTypes.DATAFRAME, 'outputDf')
class _BinaryTransformer(Stage):
    pass


@input_slot('input_df', MessageTypes.DATAFRAME, 'inputDf')
@output_slot('model', MessageTypes.MODEL)
@output_slot('output_df', MessageTypes.DATAFRAME, 'outputDf', is_default=True)
class _Estimator(Stage):
    pass


@input_slot('input_col', MessageTypes.VALUE, 'inputCol')
class _HasInputCol(Stage):
    pass


@input_slot('input_cols', MessageTypes.value(value_type='array'), 'inputCols')
class _HasInputCols(Stage):
    pass


@input_slot('output_col', MessageTypes.VALUE, 'outputCol')
class _HasOutputCol(Stage):
    pass


@input_slot('input_df', MessageTypes.DATAFRAME, 'inputDf')
@output_slot('metric_value', MessageTypes.value(value_type='double'), 'metricValue', is_default=True)
class _Evaluator(Stage):
    pass

####################################################################################

####################################################################################


class Placeholder(Stage):
    """
    General parent class for all placeholders.
    Mostly for coding convenience, not actually used in the APIs.
    """
    pass


@input_slot('input_val', MessageTypes.VALUE, server_name='inputVal')
@output_slot('output_val', MessageTypes.VALUE, server_name='outputVal')
class ValuePlaceholder(Placeholder):
    def __init__(self, name='Stage', value_type=None):
        # overwrite the actual input slot and output slot, to incorporate custom `value_type`
        msg_type = MessageTypes.VALUE if value_type is None else MessageTypes.value(value_type=value_type)
        self._input_val_slot = _Slot('input_val', msg_type, True, 'inputVal', False)
        self._output_val_slot = _Slot('output_val', msg_type, False, 'outputVal', True)

        super(ValuePlaceholder, self).__init__(name=name)

    def slot(self, slot_name='', is_input=True):
        # override this function to return the custom slots created in the constructor
        # this is to order to take into account custom `value_type` and make sure serialization works
        # especially for tricky types like list and so on.
        if slot_name == 'input_val':
            return self._input_val_slot
        if slot_name == 'output_val':
            return self._output_val_slot
        return super(ValuePlaceholder, self).slot(slot_name=slot_name, is_input=is_input)


@input_slot('input_val', MessageTypes.DATAFRAME, server_name='inputVal')
@output_slot('output_val', MessageTypes.DATAFRAME, server_name='outputVal')
class DataframePlaceholder(Placeholder):
    pass


@input_slot('input_val', MessageTypes.COLUMN, server_name='inputVal')
@output_slot('output_val', MessageTypes.COLUMN, server_name='outputVal')
class ColumnPlaceholder(Placeholder):
    pass


"""
ETL stages
"""


@input_slot('col_names', MessageTypes.value(value_type='array'), server_name='colNames')
class Drop(_UnaryTransformer):
    pass


"""
Feature extractors
"""


@input_slot('labels', MessageTypes.value(value_type='array'))
class IndexToString(_UnaryTransformer, _HasInputCol, _HasOutputCol):
    pass


@output_slot('model', MessageTypes.MODEL)
class StringIndexer(_UnaryTransformer, _HasInputCol, _HasOutputCol):
    pass


class VectorAssembler(_UnaryTransformer, _HasInputCols, _HasOutputCol):
    pass


"""
ML stages
"""


@input_slot('features_col', MessageTypes.VALUE, server_name='featuresCol')
class _HasFeaturesCol(Stage):
    pass


@input_slot('label_col', MessageTypes.VALUE, server_name='labelCol')
class _HasLabelCol(Stage):
    pass


@input_slot('prediction_col', MessageTypes.VALUE, server_name='predictionCol')
class _HasPredictionCol(Stage):
    pass


@input_slot('probability_col', MessageTypes.VALUE, server_name='probabilityCol')
class _HasProbabilityCol(Stage):
    pass


@input_slot('aggregation_depth', MessageTypes.VALUE, server_name='aggregationDepth')
@input_slot('elastic_net_param', MessageTypes.value(value_type='double'), server_name='elasticNetParam')
@input_slot('fit_intercept', MessageTypes.VALUE, server_name='fitIntercept')
@input_slot('max_iter', MessageTypes.VALUE, server_name='maxIter')
@input_slot('reg_param', MessageTypes.value(value_type='double'), server_name='regParam')
@input_slot('standardization', MessageTypes.VALUE, server_name='standardization')
@input_slot('tolerance', MessageTypes.value(value_type='double'), server_name='tolerance')
@input_slot('weight_col', MessageTypes.VALUE, server_name='weightCol')
@input_slot('solver', MessageTypes.VALUE, server_name='solver')
class _LinearRegressionInputs(_HasFeaturesCol, _HasLabelCol, _HasPredictionCol):
    pass


class LinearRegression(_Estimator, _LinearRegressionInputs):
    pass


"""
Evaluators
"""


@input_slot('metric_name', MessageTypes.VALUE, server_name='metricName')
class RegressionEvaluator(_Evaluator, _HasLabelCol, _HasPredictionCol):
    pass



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

import os
import time

import docker
import requests
import six
from docker import errors as docker_errors
from future import utils
from requests import exceptions as requests_exceptions

from pycebes import config
from pycebes.internal import helpers as pycebes_helper


def _get_docker_client():
    """Returns the docker client"""
    try:
        return docker.from_env(version='auto')
    except docker_errors.DockerException as e:
        err_msg = 'Could not create docker client: {}. Have you started the docker daemon?'.format(e)
        utils.raise_from(ValueError(err_msg), e)


@six.python_2_unicode_compatible
class _CebesContainerInfo(object):

    def __init__(self, name='', image='', container_ip='', cebes_port=0):
        self.name = name
        self.image = image
        self.container_ip = container_ip
        self.cebes_port = cebes_port
        self._logger = pycebes_helper.get_logger(self.__class__.__module__ + '.' + self.__class__.__name__)

    def __str__(self):
        return '{}[{}] at port {}'.format(self.name, self.image, self.cebes_port)

    def shutdown(self):
        """Shutdown this container."""
        client = _get_docker_client()
        try:
            container = client.containers.get(self.name)
            container.stop()
            self._logger.info('Cebes container stopped: {}'.format(self))
        except docker_errors.NotFound:
            self._logger.error('Failed to stop container {}'.format(self), exc_info=1)
        except docker_errors.APIError:
            self._logger.error('Failed to stop container {}'.format(self), exc_info=1)
        finally:
            client.api.close()


@six.python_2_unicode_compatible
class _CebesHttpServerContainerInfo(_CebesContainerInfo):
    def __init__(self, name='', image='', container_ip='', cebes_port=0, spark_port=0):
        super(_CebesHttpServerContainerInfo, self).__init__(
            name=name, image=image, container_ip=container_ip, cebes_port=cebes_port)
        self.spark_port = spark_port


class _DockerContainerStarter(object):
    """
    A helper for starting a new docker container, or return
    its details if it is running (based on its label)
    """

    def __init__(self, img_repo='phvu/cebes', img_tag='0.10.0',
                 container_name_prefix='cebes-http-server',
                 container_data_dir='/cebes/data',
                 data_sub_dir=''):
        self._logger = pycebes_helper.get_logger(self.__class__.__module__ + '.' + self.__class__.__name__)
        self._img_repo = img_repo
        self._img_tag = img_tag
        self._container_name_prefix = container_name_prefix
        self._container_data_dir = container_data_dir
        self._data_sub_dir = data_sub_dir

    def start(self):
        client = _get_docker_client()
        container_info = self._find_running_container(client)
        if container_info is None:
            container_info = self._start_cebes_container(client)

        client.api.close()
        return container_info

    def _port_mapping(self):
        return {'21000/tcp': None, '4040/tcp': None}

    def _parse_container_attrs(self, attrs):
        return _CebesHttpServerContainerInfo(
            name=attrs['Name'][1:], image=attrs['Config']['Image'],
            container_ip=attrs['NetworkSettings']['IPAddress'],
            cebes_port=int(attrs['NetworkSettings']['Ports']['21000/tcp'][0]['HostPort']),
            spark_port=int(attrs['NetworkSettings']['Ports']['4040/tcp'][0]['HostPort']))

    def _find_running_container(self, client):
        """Find all running containers having the same label."""
        running_containers = []
        for c in client.containers.list(filters={'label': 'type={}'.format(self._container_name_prefix)}):
            container_info = self._parse_container_attrs(c.attrs)
            if container_info.image == '{}:{}'.format(self._img_repo, self._img_tag):
                running_containers.append(container_info)

        if len(running_containers) > 1:
            self._logger.info('Detected multiple containers: {}'.format(
                ', '.join('{}'.format(c) for c in running_containers)))

        if len(running_containers) > 0:
            return running_containers[0]
        return None

    def _start_cebes_container(self, client):
        """
        Start a new Cebes container

        :type client: DockerClient
        :rtype: _CebesContainerInfo
        """

        # data dir
        data_path = os.path.join(os.path.expanduser('~/.cebes'), self._img_tag)
        if self._data_sub_dir:
            data_path = os.path.join(data_path, self._data_sub_dir)
        os.makedirs(data_path, mode=0o700, exist_ok=True)

        # invent a name
        name_template = '{}-{}-{{}}'.format(self._container_name_prefix, self._img_tag)
        i = 0
        container_name = name_template.format(i)
        while True:
            running_containers = client.containers.list(all=True, filters={'name': container_name})

            # if no container has the same name, then we are good
            if len(running_containers) == 0:
                break

            # if there is a container with the same name but it exited, remove it and use the name
            container = running_containers[0]
            if container.status == 'exited':
                container.remove()
                break

            # increase the id and get a new name
            i += 1
            container_name = name_template.format(i)

        self._logger.info('Starting container {}[{}:{}] with data path at {}'.format(
            container_name, self._img_repo, self._img_tag, data_path))
        self._logger.info('This may take several minutes if Cebes needs to pull the image')
        self._logger.info('If it takes too long, run "docker pull {}:{}" in a different console and try again'.format(
            self._img_repo, self._img_tag))

        container = client.containers.run('{}:{}'.format(self._img_repo, self._img_tag),
                                          detach=True, ports=self._port_mapping(),
                                          volumes={data_path: {'bind': self._container_data_dir, 'mode': 'rw'}},
                                          labels={'type': self._container_name_prefix},
                                          name=container_name)

        while container.status == 'created':
            time.sleep(0.1)
            container.reload()

        if container.status != 'running':
            self._logger.warning('Container log:\n{}'.format(container.logs().decode('utf-8')))
            raise ValueError('Unable to launch Cebes container. See logs above for more information')

        container_info = self._parse_container_attrs(container.attrs)

        # wait until the service is up
        c = 0
        max_tries = 200
        sess = requests.Session()
        connected = False
        while c < max_tries and not connected:
            try:
                sess.get('http://localhost:{}'.format(container_info.cebes_port))
                connected = True
            except requests_exceptions.ConnectionError:
                time.sleep(0.5)
                c += 1
        sess.close()

        if connected:
            self._logger.info('Cebes container started, listening at localhost:{}'.format(container_info.cebes_port))
        else:
            self._logger.warning('Cebes container {} takes more time than usual to start. '
                                 'You might want to wait a bit more before trying again'.format(container_info))
        return container_info


class _RepositoryContainerStarter(_DockerContainerStarter):
    def __init__(self, img_repo='phvu/cebes', img_tag='0.10.0',
                 container_name_prefix='cebes-http-server',
                 container_data_dir='/cebes/data',
                 data_sub_dir='',
                 host_port=None):
        super(_RepositoryContainerStarter, self).__init__(
            img_repo=img_repo, img_tag=img_tag,
            container_name_prefix=container_name_prefix,
            container_data_dir=container_data_dir,
            data_sub_dir=data_sub_dir)
        self._host_port = host_port

    def _port_mapping(self):
        return {'22000/tcp': self._host_port}

    def _parse_container_attrs(self, attrs):
        return _CebesHttpServerContainerInfo(
            name=attrs['Name'][1:], image=attrs['Config']['Image'],
            container_ip=attrs['NetworkSettings']['IPAddress'],
            cebes_port=int(attrs['NetworkSettings']['Ports']['22000/tcp'][0]['HostPort']),
            spark_port=0)


def get_cebes_http_server_container():
    """
    Get the details of a running Cebes HTTP server container,
    or start a new container if none is running
    """
    starter = _DockerContainerStarter(config.LOCAL_HTTP_SERVER_REPO, config.LOCAL_HTTP_SERVER_TAG,
                                      container_name_prefix='cebes-http-server',
                                      container_data_dir='/cebes/data',
                                      data_sub_dir='http-server')
    return starter.start()


def get_cebes_repository_container(host_port=None):
    """
    Get the details of a running Cebes repository container, or start a new one.
    """
    starter = _RepositoryContainerStarter(config.LOCAL_REPOSITORY_REPO, config.LOCAL_REPOSITORY_TAG,
                                          container_name_prefix='cebes-pipeline-repository',
                                          container_data_dir='/cebes/repository',
                                          data_sub_dir='repository',
                                          host_port=host_port)
    return starter.start()



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from pycebes.internal.default_stack import DefaultStack

_default_session_stack = DefaultStack()
_default_pipeline_stack = DefaultStack()


def _get_default(stack, error_msg='No default object found'):
    ret = stack.get_default()
    if ret is None:
        raise LookupError(error_msg)
    return ret


def get_default_session():
    """
    Get the current default session, raise Exception if there is no Session created

    :rtype: pycebes.core.session.Session
    """
    return _get_default(_default_session_stack, 'No default session found. You need to create a Session')


def get_session_stack():
    """
    Get the default session stack

    :rtype: DefaultStack
    """
    return _default_session_stack


def get_default_pipeline():
    """
    Get the current pipeline, raise Exception if there is no Pipeline created

    :rtype: pycebes.core.pipeline.Pipeline
    """
    return _get_default(_default_pipeline_stack, 'No pipeline found. You need to create a Pipeline')


def get_pipeline_stack():
    """
    Get the pipeline stack

    :rtype: DefaultStack
    """
    return _default_pipeline_stack



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import datetime
import enum

import six
from pycebes.internal.helpers import require


@enum.unique
class VariableTypes(enum.Enum):
    DISCRETE = 'Discrete'
    CONTINUOUS = 'Continuous'
    NOMINAL = 'Nominal'
    ORDINAL = 'Ordinal'
    TEXT = 'Text'
    DATETIME = 'DateTime'
    ARRAY = 'Array'
    MAP = 'Map'
    STRUCT = 'Struct'

    @classmethod
    def from_str(cls, s):
        v = next((e for e in cls.__members__.values() if e.value == s), None)
        if v is None:
            raise ValueError('Unknown variable type: {!r}'.format(s))
        return v

    def to_json(self):
        """
        Return the JSON representation of this variable type
        :return:
        """
        return self.value


class StorageType(object):
    def __init__(self, cebes_type, python_type, name=None):
        self._name = name
        self._cebes_type = cebes_type
        self._python_type = python_type

    @property
    def name(self):
        return self._name or self._cebes_type

    @property
    def cebes_type(self):
        return self._cebes_type

    @property
    def python_type(self):
        return self._python_type

    def to_json(self):
        return self._cebes_type


class ArrayType(StorageType):

    def __init__(self, element_type):
        """
        # Arguments
        element_type (StorageType): type of the element in this array
        """
        super(ArrayType, self).__init__('Array[{}]'.format(element_type.cebes_type), list,
                                        name='Array[{}]'.format(element_type.name))
        self._element_type = element_type

    @property
    def element_type(self):
        return self._element_type

    def to_json(self):
        return {'elementType': self._element_type.to_json()}

    @staticmethod
    def from_json(js_val):
        require(isinstance(js_val, dict) and 'elementType' in js_val)
        return ArrayType(StorageTypes.from_json(js_val['elementType']))


class MapType(StorageType):

    def __init__(self, key_type, value_type):
        """
        # Arguments
        key_type (StorageType): type of the key
        value_type (StorageType): type of the value
        """
        super(MapType, self).__init__('Map[{}, {}]'.format(key_type.cebes_type, value_type.cebes_type), dict,
                                      name='Map[{}, {}]'.format(key_type.name, value_type.name))
        self._key_type = key_type
        self._value_type = value_type

    @property
    def key_type(self):
        return self._key_type

    @property
    def value_type(self):
        return self._value_type

    def to_json(self):
        return {'keyType': self._key_type.to_json(), 'valueType': self._value_type.to_json()}

    @staticmethod
    def from_json(js_val):
        require(isinstance(js_val, dict) and 'keyType' in js_val and 'valueType' in js_val)
        return MapType(StorageTypes.from_json(js_val['keyType']),
                       StorageTypes.from_json(js_val['valueType']))


class StructField(object):
    def __init__(self, name, storage_type):
        self._name = name
        self._storage_type = storage_type

    @property
    def name(self):
        return self._name

    @property
    def storage_type(self):
        return self._storage_type

    def to_json(self):
        return {'name': self._name, 'storageType': self._storage_type.to_json(), 'metadata': {}}

    @staticmethod
    def from_json(js_val):
        require(isinstance(js_val, dict) and 'name' in js_val and 'storageType' in js_val)
        return StructField(js_val['name'],
                           StorageTypes.from_json(js_val['storageType']))


class StructType(StorageType):

    def __init__(self, fields):
        super(StructType, self).__init__(
            'Struct[{}]'.format(','.join(f.storage_type.cebes_type for f in fields)),
            dict,
            name='Struct[{}]'.format(','.join(f.storage_type.name for f in fields)))
        self._fields = fields

    @property
    def fields(self):
        return self._fields[:]

    def to_json(self):
        return {'fields': [f.to_json() for f in self._fields]}

    @staticmethod
    def from_json(js_val):
        require(isinstance(js_val, dict) and 'fields' in js_val)
        return StructType([StructField.from_json(f) for f in js_val['fields']])


class StorageTypes(object):
    STRING = StorageType('string', six.text_type, name='STRING')
    BINARY = StorageType('binary', six.binary_type, name='BINARY')

    DATE = StorageType('date', datetime.date, name='DATE')
    TIMESTAMP = StorageType('timestamp', int, name='TIMESTAMP')
    CALENDAR_INTERVAL = StorageType('calendarinterval', datetime.timedelta, name='CALENDAR_INTERVAL')

    BOOLEAN = StorageType('boolean', bool, name='BOOLEAN')
    SHORT = StorageType('short', int, name='SHORT')
    INTEGER = StorageType('integer', int, name='INTEGER')
    LONG = StorageType('long', int, name='LONG')
    FLOAT = StorageType('float', float, name='FLOAT')
    DOUBLE = StorageType('double', float, name='DOUBLE')
    VECTOR = StorageType('vector', list, name='VECTOR')

    __atomic_types__ = [STRING, BINARY,
                        DATE, TIMESTAMP, CALENDAR_INTERVAL,
                        BOOLEAN, SHORT, INTEGER, LONG, FLOAT, DOUBLE, VECTOR]

    @staticmethod
    def array(element_type):
        """
        # Arguments
        element_type (StorageType):
        """
        return ArrayType(element_type)

    @staticmethod
    def map(key_type, value_type):
        return MapType(key_type, value_type)

    @staticmethod
    def struct(fields):
        return StructType(fields)

    @classmethod
    def from_json(cls, js_val):
        v = next((e for e in cls.__atomic_types__ if e.cebes_type == js_val), None)
        if v is not None:
            return v

        for clz in [ArrayType, MapType, StructType]:
            try:
                return clz.from_json(js_val)
            except ValueError:
                pass
        raise ValueError('Unknown storage type: {!r}'.format(js_val))


@six.python_2_unicode_compatible
class SchemaField(object):
    def __init__(self, name='', storage_type=StorageTypes.STRING, variable_type=VariableTypes.TEXT):
        self.name = name
        self.storage_type = storage_type
        self.variable_type = variable_type

    def __repr__(self):
        return '{}(name={!r},storage_type={},variable_type={})'.format(
            self.__class__.__name__, self.name, self.storage_type.name, self.variable_type.name)


@six.python_2_unicode_compatible
class Schema(object):
    def __init__(self, fields=None):
        self.fields = fields

    @property
    def columns(self):
        """A list of column names in this Schema"""
        return [f.name for f in self.fields]

    @property
    def simple_string(self):
        """Return a simple description of this schema, mainly used for brevity"""
        n = 3
        if len(self.fields) <= 0:
            s = '(empty)'
        else:
            s = ', '.join('{} {}'.format(f.name, f.storage_type.cebes_type) for f in self.fields[:n])
            if len(self.fields) > n:
                s += '... (+{} columns)'.format(len(self.fields) - n)
        return s

    def __repr__(self):
        return '{}(fields=[{}])'.format(
            self.__class__.__name__, ','.join('{!r}'.format(f) for f in self.fields))

    def __len__(self):
        return len(self.fields)

    def __getitem__(self, item):
        try:
            return next(f for f in self.fields if f.name == item)
        except StopIteration:
            raise KeyError('Column not found: {!r}'.format(item))

    @classmethod
    def from_json(cls, js_data):
        """
        Parse the Schema from its JSON representation

        :param js_data: a dict with a key named ``fields``, 
            which is a list of dict, containing the schema fields. Each field is a dict.
        :type js_data: dict
        :rtype: Schema
        """
        fields = [SchemaField(name=d.get('name', ''),
                              storage_type=StorageTypes.from_json(d.get('storageType', '')),
                              variable_type=VariableTypes.from_str(d.get('variableType', '')))
                  for d in js_data.get('fields', [])]
        return Schema(fields=fields)



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import six

from pycebes.core import dataframe
from pycebes.core.stages import SlotDescriptor, Stage, MessageType, Placeholder
from pycebes.internal.helpers import require
from pycebes.internal.implicits import get_pipeline_stack, get_default_session


class Model(object):
    """
    Represent a trained model
    """

    def __init__(self, _id, model_class, inputs, metadata):
        self._id = _id
        self._model_class = model_class
        self._metadata = metadata
        self._inputs = {}
        for k, v in inputs.items():
            param_name = ''.join('_{}'.format(c.lower()) if c.isupper() else c for c in k)
            self._inputs[param_name] = v

    def __repr__(self):
        return '{}(id={!r},model_class={!r})'.format(self.__class__.__name__, self.id, self._model_class)

    @property
    def id(self):
        return self._id

    @property
    def metadata(self):
        return dict(**self._metadata)

    @property
    def inputs(self):
        return dict(**self._inputs)

    def transform(self, input_df):
        """
        Transform the given Dataframe

        # Arguments
        input_df (Dataframe): the input Dataframe to be transformed

        # Returns
        Dataframe: the Dataframe transformed by this model
        """
        data = {'model': {'modelId': self._id},
                'inputDf': {'dfId': input_df.id}}
        result = get_default_session().client.post_and_wait('model/run', data)
        return dataframe.Dataframe.from_json(result)

    @classmethod
    def from_json(cls, js_data):
        """
        Deserialize a Model received from the server

        :param js_data: JSON data received from server
        :rtype: Model
        """
        inputs = {}
        for k, v in js_data.get('inputs', {}).items():
            inputs[k] = MessageType.from_json(v)
        return Model(_id=js_data['id'], model_class=js_data['modelClass'], inputs=inputs,
                     metadata=js_data.get('metaData', {}))


@six.python_2_unicode_compatible
class Pipeline(object):
    def __init__(self, _id=None, stages=None):
        self._id = _id
        self._stages = [] if stages is None else stages
        self._context_manager = None

    @property
    def id(self):
        return self._id

    @property
    def stages(self):
        """Convenient method to list all stages in a dict, indexed by their name"""
        return {s.get_input(s.name): s for s in self._stages}

    def __getitem__(self, item):
        try:
            return next(s for s in self._stages if s.get_name() == item)
        except StopIteration:
            raise KeyError('Stage not found: {!r}'.format(item))

    def __contains__(self, item):
        """
        Check if the given stage or stage name belong to this pipeline

        :param item: stage name (string) or a stage object
        :return: true if the given stage belong to this pipeline
        """
        if isinstance(item, Stage):
            return next((s for s in self._stages if s == item), None) is not None

        require(isinstance(item, six.text_type), 'Only support stage object or a string, got {!r}'.format(item))
        return next((s for s in self._stages if s.get_name() == item), None) is not None

    def __enter__(self):
        self._context_manager = get_pipeline_stack().get_controller(self)
        return self._context_manager.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._context_manager.__exit__(exc_type, exc_val, exc_tb)

    def __repr__(self):
        stage_names = ', '.join(s.get_input(s.name) for s in self._stages)
        return '{}({})'.format(self.__class__.__name__, stage_names)

    def add(self, stage):
        """
        Add a stage into this pipeline. No-op if the stage is already in this pipeline.

        Note: the stage must have a name before it is added into a Pipeline.
        This function throws exception if the stage doesn't have a name, or it is given a duplicated name

        :type stage: Stage
        :return: the given stage
        """
        require(self._id is None, 'Cannot add stage into this Pipeline since it is already created on the server')

        if stage not in self._stages:
            stage_name = stage.get_name()
            require(stage_name is not None, 'Please give the stage a name before adding it into the pipeline')
            require(next((s for s in self._stages if s.get_name() == stage_name), None) is None,
                    'Duplicated stage name: {}'.format(stage.get_name()))
            self._stages.append(stage)
        return stage

    @classmethod
    def from_json(cls, js_data):
        """
        Return a ``Pipeline`` instance from its JSON representation

        :param js_data: a dict with ``id`` and ``stages``
        :rtype: Pipeline
        """
        stages = [Stage.from_json(js) for js in js_data['stages']]

        def _get_stage(stage_name):
            """
            :rtype: Stage
            """
            return next(s for s in stages if s.get_name() == stage_name)

        # dirty: scan the stages again and fill-in the SlotDescriptors
        for js_stage in js_data['stages']:
            this_stage = _get_stage(js_stage['name'])
            for k, v in js_stage['inputs'].items():

                if v[0] == MessageType.STAGE_OUTPUT_DEF:
                    # ['StageOutputDef', {'stageName': 'name-of-the-stage', 'outputName': 'outputVal'}]
                    parent_stage = _get_stage(v[1]['stageName'])
                    slot_desc = parent_stage.slot_descriptor(v[1]['outputName'])

                    this_stage._set_input(this_stage.slot_descriptor(k), slot_desc)

        return Pipeline(js_data['id'], stages=stages)

    def to_json(self):
        """
        Return the JSON representation of this Pipeline
        """
        return {'id': self.id, 'stages': [s.to_json() for s in self._stages]}

    def run(self, outputs=(), feeds=None, timeout=-1):
        """
        Run this pipeline, given the feeds and take the outputs

        :param outputs: list of SlotDescriptor from which to take the value
        :param feeds: a dictionary of {SlotDescriptor -> value} giving the values
            to some slots in the pipeline
        :param timeout: timeout in seconds. Negative means wait indefinitely
        :return: tuple of values of the output slots
        """
        single_output = False
        if not isinstance(outputs, (tuple, list)):
            single_output = True
            outputs = [outputs]
        output_slots = []

        for o in outputs:
            # convenience: automatically translate stages into their corresponding default output slot descriptor
            slot_desc = o
            if isinstance(slot_desc, Stage):
                slot_desc = slot_desc.default_output_slot_descriptor
            require(isinstance(slot_desc, SlotDescriptor), 'Expected an output slot, got {!r}'.format(o))
            output_slots.append(slot_desc.to_json())

        feeds = feeds or {}
        feeds_json = {}
        for k, v in feeds.items():
            # convenience: automatically translate placeholders into its input slot
            slot_desc = k
            if isinstance(slot_desc, Placeholder):
                slot_desc = slot_desc.input_val

            require(isinstance(slot_desc, SlotDescriptor),
                    'Expected slots as keys in `feeds`, got {!r}'.format(slot_desc))
            require(slot_desc.message_type.is_valid(v), 'Invalid value {!r} for slot {!r}'.format(v, slot_desc))

            feeds_json[slot_desc.full_server_name] = slot_desc.message_type.to_json(v)

        ppl_json = self.to_json()

        data = {'pipeline': ppl_json,
                'feeds': feeds_json,
                'outputs': output_slots,
                'timeout': timeout}

        run_result = get_default_session().client.post_and_wait('pipeline/run', data)

        assert self._id is None or self._id == run_result['pipelineId']
        self._id = run_result['pipelineId']

        # parse the results
        require(len(run_result['results']) == len(output_slots), 'Invalid result from server')
        parsed_results = []
        for out_slot in output_slots:
            r = next((v for k, v in run_result['results'] if k == out_slot), None)
            require(r is not None, 'Could not find result for output slot {}'.format(out_slot))
            parsed_results.append(MessageType.from_json(r))

        return parsed_results[0] if single_output else tuple(parsed_results)



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import


class ServerException(Exception):
    """
    Exception happened on the server, got catched and returned to the client
    """

    def __init__(self, message='', server_stack_trace='', request_uri='', request_entity=None):
        super(ServerException, self).__init__(message, server_stack_trace, request_uri, request_entity)
        self.server_stack_trace = server_stack_trace
        self.request_uri = request_uri
        self.request_entity = request_entity



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import six

import pycebes.core.expressions as exprs
from pycebes.core.schema import StorageTypes
from pycebes.internal.helpers import require


def lit(literal):
    """
    Return a Column object containing the given literal
    """
    if isinstance(literal, Column):
        return literal
    return Column(exprs.Literal(literal))


@six.python_2_unicode_compatible
class Column(object):
    def __init__(self, expr):
        """
        Represent a column of a #Dataframe, backed by an expression.

        # Arguments
        expr (exprs.Expression): Expression behind this column
        """
        require(isinstance(expr, exprs.Expression), 'expr has to be an Expression, got {!r}'.format(expr))
        self.expr = expr

    def __repr__(self):
        return '{}(expr={!r})'.format(self.__class__.__name__, self.expr)

    """
    helpers
    """

    def _with_expr(self, expr_class, *args):
        """
        Helper to return a new column with the given expression class, constructed
        with the current expression as the first argument, and
        other arguments in *args
        """
        return Column(expr_class(self.expr, *args))

    def _bin_op(self, expr_class, other):
        """
        Binary operation with this expression (self.expr) and the other (functions.lit(other).expr)
        """
        return Column(expr_class(self.expr, lit(other).expr))

    def to_json(self):
        """
        Serialize the Column into JSON format

        # Returns
         JSON representation of this Column
        """
        return {'expr': self.expr.to_json()}

    """
    Building expressions
    """

    @property
    def desc(self):
        """
        Returns an ordering used in sorting.

        # Example
        ```python
        df.sort(df.col1.desc)
        ```
        """
        return self._with_expr(exprs.SortOrder, exprs.SortOrder.Descending)

    @property
    def asc(self):
        """
        Returns an ordering used in sorting.

        # Example
        ```python
        df.sort(df.col1.asc)
        ```
        """
        return self._with_expr(exprs.SortOrder, exprs.SortOrder.Ascending)

    def __neg__(self):
        """
        Unary minus, i.e. negate the expression.

        # Example
        ```python
        df.select(-df.cost)
        ```
        """
        return self._with_expr(exprs.UnaryMinus)

    def __nonzero__(self):
        raise ValueError("Cannot convert column into bool: please use '&' for 'and', '|' for 'or', "
                         "'~' for 'not' when building Dataframe boolean expressions.")

    __bool__ = __nonzero__

    def __invert__(self):
        """
        Inversion of boolean expression, i.e. NOT.

        # Example
        ```python
        # select rows that are not rich (isRich === false)
        df.filter(~(df.isRich))
        ```
        """
        return self._with_expr(exprs.Not)

    def __eq__(self, other):
        """
        Equality test.

        # Example
        ```python
        df.filter(df.colA == df.colB)
        ```
        """
        return self._bin_op(exprs.EqualTo, other)

    def __ne__(self, other):
        """
        Inequality test.

        # Example
        ```python
        df.select(df.colA != df.colB)
        ```
        """
        return Column(exprs.Not(self.__eq__(other)))

    def __gt__(self, other):
        """
        # Example
        ```python
        # The following selects people older than 25.
        people.select(people.age > 25)
        ```
        """
        return self._bin_op(exprs.GreaterThan, other)

    def __lt__(self, other):
        """
        # Example
        ```python
        # The following selects people younger than 25.
        people.select(people.age < 25)
        ```
        """
        return self._bin_op(exprs.LessThan, other)

    def __le__(self, other):
        """
        # Example
        ```python
        # The following selects people younger than 25.
        people.select(people.age <= 25)
        ```
        """
        return self._bin_op(exprs.LessThanOrEqual, other)

    def __ge__(self, other):
        """
        # Example
        ```python
        # The following selects people younger than 25.
        people.select(people.age >= 25)
        ```
        """
        return self._bin_op(exprs.GreaterThanOrEqual, other)

    def equal_null_safe(self, other):
        """
        Equality test that is safe for null values.
        Returns same result with EQUAL(=) operator for non-null operands,
        but returns ``true`` if both are ``NULL``, ``false`` if one of the them is ``NULL``.
        """
        return self._bin_op(exprs.EqualNullSafe, other)

    def when(self, condition, value):
        """
        Evaluates a list of conditions and returns one of multiple possible result expressions.
        If otherwise() is not defined at the end, `null` is returned for unmatched conditions.

        # Arguments
        condition: the condition to be checked
        value (Column): the output value if condition is true

        # Example
        ```python
        # encoding gender string column into integer.
        people.select(functions.when(people.gender == "male", 0)
            .when(people.gender == "female", 1)
            .otherwise(2))
        ```
        """
        if not isinstance(condition, Column):
            raise ValueError('Expect condition as a Column')
        if not isinstance(self.expr, exprs.CaseWhen):
            raise ValueError('when() can only be applied on a Column previously generated by when() function')
        if self.expr.else_value is not None:
            raise ValueError('when() cannot be applied once otherwise() is applied')
        return Column(exprs.CaseWhen(self.expr[:] + [(condition, lit(value).expr)]))

    def otherwise(self, value):
        """
        Evaluates a list of conditions and returns one of multiple possible result expressions.
        If otherwise() is not defined at the end, `null` is returned for unmatched conditions.

        # Example
        ```python
        # encoding gender string column into integer.
        people.select(functions.when(people.gender == "male", 0)
            .when(people.gender == "female", 1)
            .otherwise(2))
        ```
        """
        if not isinstance(self.expr, exprs.CaseWhen):
            raise ValueError('otherwise() can only be applied on a Column previously generated by when() function')
        if self.expr.else_value is not None:
            raise ValueError('otherwise() can only be applied once on a Column previously generated by when()')
        return Column(exprs.CaseWhen(self.expr[:], lit(value).expr))

    def between(self, lower_bound, upper_bound):
        """
        True if the current column is between the lower bound and upper bound, inclusive.
        """
        return (self >= lower_bound) & (self <= upper_bound)

    def is_nan(self):
        """
        True if the current expression is NaN.
        """
        return self._with_expr(exprs.IsNaN)

    def is_null(self):
        """
        True if the current expression is null.
        """
        return self._with_expr(exprs.IsNull)

    def is_not_null(self):
        """
        True if the current expression is not null.
        """
        return self._with_expr(exprs.IsNotNull)

    def __or__(self, other):
        """
        Boolean OR.

        # Example
        ```python
        # Selects people that are in school or employed.
        people.filter(people.inSchool | people.isEmployed)
        ```
        """
        return self._bin_op(exprs.Or, other)

    def __and__(self, other):
        """
        Boolean AND.

        # Example
        ```python
        # Selects people that are in school and employed at the same time.
        people.select(people.inSchool & people.isEmployed)
        ```
        """
        return self._bin_op(exprs.And, other)

    def __add__(self, other):
        """
        Sum of this expression and another expression. Only work on numeric columns.

        # Example
        ```python
        # Selects the sum of a person's height and weight.
        people.select(people.height + people.weight)
        ```
        """
        return self._bin_op(exprs.Add, other)

    def __sub__(self, other):
        """
        Subtraction. Subtract the other expression from this expression. Only work on numeric columns.

        # Example
        ```python
        # Selects the difference between people's height and their weight.
        people.select(people.height - people.weight)
        ```
        """
        return self._bin_op(exprs.Subtract, other)

    def __mul__(self, other):
        """
        Multiplication of this expression and another expression. Only work on numeric columns.

        # Example
        ```python
        # Multiplies a person's height by their weight
        people.select(people.height * people.weight)
        ```
        """
        return self._bin_op(exprs.Multiply, other)

    def __truediv__(self, other):
        """
         Division this expression by another expression. Only work on numeric columns.

        # Example
        ```python
        # Divides a person's height by their weight
        people.select(people.height / people.weight)
        ```
        """
        return self._bin_op(exprs.Divide, other)

    def __mod__(self, other):
        """
        Modulo (a.k.a. remainder) expression.
        """
        return self._bin_op(exprs.Remainder, other)

    def isin(self, values):
        """
        A boolean expression that is evaluated to true if the value of this expression is contained
        by the evaluated values of the arguments.
        """
        return self._with_expr(exprs.In, [lit(v).expr for v in values])

    def like(self, literal):
        """
        SQL like expression. Result will be a BooleanType column

        # Arguments
        literal: a string
        """
        if not isinstance(literal, six.text_type):
            raise ValueError('Expected a string, got {!r}'.format(literal))
        return self._with_expr(exprs.Like, literal)

    def rlike(self, literal):
        """
        SQL RLIKE expression (LIKE with Regex). Result will be a BooleanType column

        # Arguments
        literal: a string
        """
        if not isinstance(literal, six.text_type):
            raise ValueError('Expected a string, got {!r}'.format(literal))
        return self._with_expr(exprs.RLike, literal)

    def get_item(self, key):
        """
        An expression that gets an item at position `ordinal` out of an array,
        or gets a value by key `key` in a MapType column.
        """
        return self._bin_op(exprs.GetItem, key)

    def get_field(self, field_name='key'):
        """
        An expression that gets a field by name in a StructType column.
        """
        return self._bin_op(exprs.GetField, field_name)

    def substr(self, start_pos, length):
        """
        An expression that returns a substring.

        This is not zero-based, but 1-based index. The first character in str has index 1.

        `start_pos` and `length` are handled specially:

          - ``"Content".substr(1, 3)`` gives ``"Con"``
          - ``"Content".substr(-100, 2)`` gives ``""``
          - ``"Content".substr(-100, 102)`` gives ``"Content"``
          - ``"Content".substr(2, 100)`` gives ``"ontent"``

        # Arguments
        start_pos: starting position, can be an integer, or a ``Column`` that gives an integer
        length: length of the sub-string, can be an integer, or a ``Column`` that gives an integer
        """
        return self._with_expr(exprs.Substring, lit(start_pos).expr, lit(length).expr)

    def contains(self, other):
        """
        Contains the other element
        """
        return self._bin_op(exprs.Contains, other)

    def starts_with(self, other):
        """
        String starts with.

        # Arguments
        other: string to test, can be a string, or a ``Column``  that gives a string
        """
        return self._bin_op(exprs.StartsWith, other)

    def ends_with(self, other):
        """
        String ends with.

        # Arguments
        other: string to test, can be a string, or a ``Column``  that gives a string
        """
        return self._bin_op(exprs.StartsWith, other)

    def alias(self, *alias):
        """
        Gives the column an alias, or multiple aliases in case it is used on an expression
        that returns more than one column (such as ``explode``).

        # Example
        ```python
        # Renames colA to colB in select output.
        df.select(df.colA.alias('colB'))

        # multiple alias
        df.select(functions.explode(df.myMap).alias('key', 'value'))
        ```
        """
        if len(alias) == 1:
            return self._with_expr(exprs.Alias, alias[0])
        return self._with_expr(exprs.MultiAlias, alias)

    def name(self, *alias):
        """
        Gives the column a name. Same as :func:`alias`
        """
        return self.alias(*alias)

    def cast(self, to):
        """
        Casts the column to a different data type.

        # Arguments
        to (StorageType): the StorageType to cast to

        # Example
        ```python
        # Casts colA to [[IntegerType]].
        df.select(df.colA.cast(IntegerType))
        df.select(df.colA.cast("int"))
        ```
        """
        if isinstance(to, six.text_type):
            to = StorageTypes.from_json(to)
        return self._bin_op(exprs.Cast, to)

    def bitwise_or(self, other):
        """
        Compute bitwise OR of this expression with another expression.

        # Example
        ```python
            df.select(df.colA.bitwise_or(df.colB))
        ```
        """
        return self._bin_op(exprs.BitwiseOr, other)

    def bitwise_and(self, other):
        """
        Compute bitwise AND of this expression with another expression.

        # Example
        ```python
        df.select(df.colA.bitwise_and(df.colB))
        ```
        """
        return self._bin_op(exprs.BitwiseAnd, other)

    def bitwise_xor(self, other):
        """
        Compute bitwise XOR of this expression with another expression.

        # Example
        ```python
        df.select(df.colA.bitwise_xor(df.colB))
        ```
        """
        return self._bin_op(exprs.BitwiseXor, other)

    @property
    def bitwise_not(self):
        """
        Compute bitwise NOT of this expression

        # Example
        ```python
        df.select(df.colA.bitwise_not)
        ```
        """
        return self._with_expr(exprs.BitwiseNot)



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import threading
import contextlib


class DefaultStack(threading.local):
    """A thread-local stack of objects for providing implicit defaults."""

    def __init__(self):
        super(DefaultStack, self).__init__()
        self._enforce_nesting = True
        self.stack = []

    def get_default(self):
        return self.stack[-1] if len(self.stack) >= 1 else None

    def reset(self):
        self.stack = []

    @property
    def enforce_nesting(self):
        return self._enforce_nesting

    @enforce_nesting.setter
    def enforce_nesting(self, value):
        self._enforce_nesting = value

    @contextlib.contextmanager
    def get_controller(self, default):
        """A context manager for manipulating a default stack."""
        try:
            self.stack.append(default)
            yield default
        finally:
            if self._enforce_nesting:
                if self.stack[-1] is not default:
                    raise AssertionError('Nesting violated for default stack of {} objects'.format(type(default)))
                self.stack.pop()
            else:
                self.stack.remove(default)



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import six
import datetime


def from_json(js):
    """
    Helper to parse json values from server into python types
    """
    if js is None or js is True or js is False or isinstance(js, six.text_type):
        # JsNull, JsBoolean, JsString
        return js

    if not isinstance(js, dict) or 'type' not in js or 'data' not in js:
        raise ValueError('Expected a dict, got {!r}'.format(js))

    t = js['type']
    data = js['data']

    if t in ('byte', 'short', 'int', 'long'):
        return int(data)

    if t in ('float', 'double'):
        return float(data)

    if t == 'timestamp':
        # server return timestamp in milliseconds, which is not the python convention
        return float(data) / 1E3

    if t == 'date':
        # server return timestamp in milliseconds
        return datetime.date.fromtimestamp(float(data) / 1E3)

    if t == 'byte_array':
        return bytearray([int(x) for x in data])

    if t in ('wrapped_array', 'seq', 'array'):
        return [from_json(x) for x in data]

    if t == 'map':
        d = {}
        for entry in data:
            if 'key' not in entry or 'val' not in entry:
                raise ValueError('Invalid map entry: {!r}'.format(entry))
            d[from_json(entry['key'])] = from_json(entry['val'])
        return d

    raise ValueError('Failed to parse value: {!r}'.format(js))


def _to_js_object(data_type='', data=None):
    """
    Seriaize the given data with the given type.
    Following "weird" convention of the server
    """
    return {'type': data_type, 'data': data}


def to_json(value, param_type=None):
    """
    Serialize (generic) python value into JSON, to be read by server
    """
    if value is None or isinstance(value, (six.text_type, bool)):
        return value

    if isinstance(value, int):
        param_type = param_type or 'int'
        assert param_type in ('byte', 'short', 'int', 'long')
        return _to_js_object(param_type, value)

    if isinstance(value, float):
        param_type = param_type or 'float'
        assert param_type in ('float', 'double', 'timestamp')
        if param_type == 'timestamp':
            # server expects timestamp in milliseconds
            value *= 1E3
        return _to_js_object(param_type, value)

    if isinstance(value, datetime.date):
        return _to_js_object('date', (value - datetime.date(1970, 1, 1)).total_seconds())

    if isinstance(value, bytearray):
        return _to_js_object('byte_array', [int(x) for x in value])

    if isinstance(value, (list, tuple)):
        param_type = param_type or 'seq'
        assert param_type in ('array', 'wrapped_array', 'seq')
        return _to_js_object(param_type, [to_json(vv) for vv in value])
    if isinstance(value, dict):
        return _to_js_object('map', [{'key': to_json(k), 'val': to_json(v)} for k, v in value.items()])

    raise ValueError('Failed to serialize value: {!r}'.format(value))



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import datetime

import tabulate
import collections
from pycebes.core.schema import Schema


class _TaggedResponseEntry(object):
    def __init__(self, js_entry):
        self.tag = js_entry['tag']
        self.id = js_entry['id']
        self.created_at = datetime.datetime.utcfromtimestamp(js_entry['createdAt'] / 1E3)

    def to_dict(self):
        result = collections.OrderedDict()
        result['UUID'] = self.id
        result['Tag'] = self.tag
        result['Created'] = self.created_at
        return result


class _TaggedDataframeResponseEntry(_TaggedResponseEntry):
    def __init__(self, js_entry):
        super(_TaggedDataframeResponseEntry, self).__init__(js_entry)
        self.schema = Schema.from_json(js_entry['schema'])

    def to_dict(self):
        d = super(_TaggedDataframeResponseEntry, self).to_dict()
        d['Schema'] = self.schema.simple_string
        return d


class _TaggedModelResponseEntry(_TaggedResponseEntry):
    def __init__(self, js_entry):
        super(_TaggedModelResponseEntry, self).__init__(js_entry)
        self.model_class = js_entry['model']['modelClass']

    def to_dict(self):
        d = super(_TaggedModelResponseEntry, self).to_dict()
        d['Model class'] = self.model_class
        return d


class _TaggedPipelineResponseEntry(_TaggedResponseEntry):
    def __init__(self, js_entry):
        super(_TaggedPipelineResponseEntry, self).__init__(js_entry)
        self.n_stages = len(js_entry['pipeline']['stages'])

    def to_dict(self):
        d = super(_TaggedPipelineResponseEntry, self).to_dict()
        d['# of stages'] = self.n_stages
        return d


"""
The response objects
"""


class _TaggedResponse(object):
    def __init__(self, js_data, tag_entry_class):
        self.tagged_objects = []
        for entry in js_data:
            self.tagged_objects.append(tag_entry_class(entry))

    def __len__(self):
        return len(self.tagged_objects)

    def __repr__(self):
        return tabulate.tabulate((e.to_dict() for e in self.tagged_objects), headers='keys')


class TaggedDataframeResponse(_TaggedResponse):
    """Result of "df/tags"""

    def __init__(self, js_data):
        super(TaggedDataframeResponse, self).__init__(js_data, _TaggedDataframeResponseEntry)


class TaggedModelResponse(_TaggedResponse):
    """Result of "model/tags"""

    def __init__(self, js_data):
        super(TaggedModelResponse, self).__init__(js_data, _TaggedModelResponseEntry)


class TaggedPipelineResponse(_TaggedResponse):
    """Result of "pipeline/tags"""

    def __init__(self, js_data):
        super(TaggedPipelineResponse, self).__init__(js_data, _TaggedPipelineResponseEntry)



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

import logging


def require(condition, msg='Requirement failed'):
    """
    Helper to raise ValueError exception if the given condition is not satisfied

    :param condition: a boolean expression to check
    :param msg: message of the exception
    :return: nothing
    """
    if not condition:
        raise ValueError(msg)


def get_logger(name):
    """
    Get a logger with default configuration and given name

    :rtype: logging.Logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    if next((c for c in logger.handlers if isinstance(c, logging.StreamHandler)), None) is None:
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        formatter = logging.Formatter('[%(levelname)-4s] %(filename)-10s: %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    return logger



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import json
import random
import time

import requests
from future.utils import raise_from
from requests import exceptions as requests_exceptions
from requests_toolbelt import MultipartEncoderMonitor

from pycebes.core.exceptions import ServerException
from pycebes.internal.helpers import require


class Client(object):
    """
    Represent a connection to the Cebes server. Normally created by a session.
    """

    def __init__(self, host='localhost', port=21000, user_name='',
                 password='', api_version='v1', interactive=True):
        self.host = host
        self.port = port
        self.user_name = user_name
        self.api_version = api_version
        self.interactive = interactive

        self.session = requests.Session()
        self.session.headers.update({'Content-Type': 'application/json'})

        self._check_server_version()

        # login
        r = self.session.post(self._server_url('auth/login'),
                              data=json.dumps({'userName': user_name, 'passwordHash': password}))
        self.session.headers.update({'Authorization': r.headers.get('Set-Authorization'),
                                     'Refresh-Token': r.headers.get('Set-Refresh-Token'),
                                     'X-XSRF-TOKEN': r.cookies.get('XSRF-TOKEN')})

    def upload(self, path):
        """
        Upload the given path to the server, return the JSON response

        :param path: path to the file to be uploaded
        :return: a dict object with 'path' and 'size'
        """

        def callback(encoder):
            if self.interactive:
                n = 20
                pct = encoder.bytes_read / encoder.len
                l = int(round(n * pct))
                s = '\rUploading: {}{} {:.0f}%'.format('.' * l, ' ' * (max(0, n - l)), pct * 100)
                print(s, end='')

        with open(path, 'rb') as f:
            monitor = MultipartEncoderMonitor.from_fields(fields={'file': f}, callback=callback)
            headers = {'Content-Type': monitor.content_type}

            response = self.session.put(self._server_url('storage/upload'), data=monitor, headers=headers)
            require(response.status_code == requests.codes.ok, 'Unsuccessful request: {}'.format(response.text))
            if self.interactive:
                print('')
            return response.json()

    def post(self, uri, data):
        """
        Send a POST request to the given uri, with the given data
        This function catches the exceptions.

        :return: a JSON object of the response
        :exception ConnectionError: if a connection to the server can't be established.
        :exception ValueError: if the response code is not OK
        """
        try:
            response = self.session.post(self._server_url(uri), data=json.dumps(data))
            require(response.status_code == requests.codes.ok, 'Unsuccessful request: {}'.format(response.text))
            return response.json()

        except requests_exceptions.ConnectionError as e:
            # wrap this in the standard ConnectionError to ease end-users
            raise_from(ConnectionError('{}'.format(e)), e)

    def wait(self, request_id, sleep_base=0.5, max_count=100):
        """
        Wait for the given request ID to complete, using an exponential back-off scheme, where:
         - wait for at most `max_count` iterations
         - at each iteration `i`: 
            - if result is ready, return
            - if not, sleep for `sleep_base * randint(0, [2 ** min(i, 7)] - 1)` seconds

        :param request_id: ID of the request to wait for
        :param sleep_base: base of 1 sleep, in seconds
        :param max_count: maximum number of iterations to wait for
        :return: the JSON object
         
        :raises TimeoutError: if ``max_count`` is exceeded
        :raises ServerException: if the request failed, e.g. an exception thrown on the server
        :raises ValueError: invalid status received from the server 
        """
        if self.interactive:
            print('Request ID: {}'.format(request_id))

        status = self.post('request/{}'.format(request_id), {})
        cnt = 0

        while status.get('status', '') == 'scheduled':
            time.sleep(sleep_base * random.randint(0, (2 ** min(cnt, 7)) - 1))
            cnt += 1
            if cnt >= max_count:
                raise TimeoutError('Timed out waiting for request ID {} after {} sleeps'.format(request_id, cnt))

            status = self.post('request/{}'.format(request_id), {})

        request_status = status.get('status', '')
        if request_status == 'finished':
            return status.get('response', {})

        if request_status == 'failed':
            fail_response = status.get('response', {})
            raise ServerException(message=fail_response.get('message', 'Unknown server exception'),
                                  server_stack_trace=fail_response.get('stackTrace', ''),
                                  request_uri=status.get('requestUri', ''),
                                  request_entity=status.get('requestEntity'))

        raise ValueError('Request ID {}: invalid status response from server: {}'.format(request_id, status.text))

    def post_and_wait(self, uri, data):
        """
        POST a request to the server, which is expected to return an ID that will be 
        used to checking the results in an exponential-backoff fashion.

        :return: a JSON object of the response
        """
        response = self.post(uri, data=data)
        request_id = response.get('requestId', None)
        require(request_id is not None, 'Request ID not found. Maybe this is not an asynchronous command? '
                                        'uri={}, response={}'.format(uri, response))
        return self.wait(request_id)

    """
    Private helpers
    """

    def _server_url(self, uri):
        return 'http://{}:{}/{}/{}'.format(self.host, self.port, self.api_version, uri)

    def _check_server_version(self):
        """
        Private helper to check if the server supports the given API version
        Raise ValueError if that is not the case.
        """
        r = self.session.get('http://{}:{}/version'.format(self.host, self.port))
        require(r.status_code == requests.codes.ok, 'Unable to query server API version: {}'.format(r.text))
        server_version = r.json()
        server_api_version = server_version.get('api', '')
        require(server_api_version == self.api_version,
                'Mismatch API version: server={}, client={}'.format(server_api_version, self.api_version))



# Copyright 2016 The Cebes Authors. All Rights Reserved.
#
# Licensed under the Apache License, version 2.0 (the "License").
# You may not use this work except in compliance with the License,
# which is available at www.apache.org/licenses/LICENSE-2.0
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied, as more fully set forth in the License.
#
# See the NOTICE file distributed with this work for information regarding copyright ownership.

from __future__ import print_function
from __future__ import unicode_literals

import os
from setuptools import setup, find_packages


def __read_requirements():
    with open(os.path.join(os.path.split(__file__)[0], 'requirements.txt'), 'r') as f:
        return [s.strip() for s in f.readlines()]


setup(
    name='pycebes',
    version='0.10.1',
    packages=find_packages(exclude=['tests']),
    description='Python client for Cebes HTTP server.',
    author='Vu Pham',
    author_email='vuph@cebes.io',
    license='Apache 2.0',
    long_description=open('README.rst', 'rb').read().decode('utf8'),
    install_requires=__read_requirements(),
)
